#   -*- coding: utf-8 -*-
#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010-2011 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Fabric commands for building a package for a direct server upgrade.

The local machine must have bzr-builder and pbuilder, and a lucid chroot
setup as the default pbuilder distribution:

    $ sudo pbuilder create --distribution lucid

"""

from __future__ import absolute_import
from __future__ import with_statement

__metaclass__ = type
__all__ = [
    'build_and_install_deb',
    'build_binary_package',
    'build_source_package',
    'clean_local_branch',
    'install_binary_package',
    'migrate_database',
    'preflight_check',
    'restart_services',
    'update_configs',
    ]

import os
import shutil
import tempfile
from datetime import datetime
from glob import glob
from textwrap import dedent

from fabric.api import (
    env,
    local,
    put,
    run,
    sudo,
    )
from fabric.contrib.files import exists


def clean_local_branch():
    local("find . -name '*.pyc' -delete")


def build_source_package(build_dir, distroseries='lucid'):
    clean_local_branch()
    # We use a recipe to ensure that we always generate later versions
    # with a timestamp. The recipe refers to the local branch '.' rather
    # than pulling from the network.
    recipe = dedent("""\
        # bzr-builder format 0.2 deb-version {debupstream}-0~{revno}+{time}
        .""")
    recipe_path = os.path.join(build_dir, 'rnr.recipe')
    with open(recipe_path, 'w+') as recipe_file:
        recipe_file.write(recipe)

    local('bzr dailydeb --distribution={0} {1} {2}'.format(
        distroseries, recipe_path, build_dir))
    dsc_file = glob(os.path.join(build_dir, '*.dsc'))[0]

    print("Created source package at {0}.".format(dsc_file))
    return dsc_file


def build_binary_package(build_dir, source_dsc):
    print("Building binary package for {0}".format(source_dsc))
    local('sudo pbuilder build --buildresult {0} {1}'.format(
        build_dir, source_dsc))
    deb_file = glob(os.path.join(build_dir, "*.deb"))[0]
    print("Created binary at {0}.".format(deb_file))
    return deb_file


def migrate_database(config_dir=None):
    if config_dir is None:
        # By default, the rnr config directory is assumed to be
        # /home/username/django_project.
        config_dir = 'django_project'
    run(config_dir + '/manage.py syncdb --migrate')


def preflight_check(config_dir='django_project'):
    # achuni 2011-03-04: Our remote user will typically not be the one that
    # runs the web server, so we need to use our own lpib cache directory.
    # This means we're not checking the launchpadlib_dir setting. :(
    run('COLUMNS=40 %s/manage.py preflight '
        '--launchpad_launchpadlib_dir=/tmp/preflight_cache' % config_dir)


def restart_services():
    sudo('service apache2 graceful', shell=False)


def install_binary_package(deb_file):
    put(deb_file, '.')
    deb_filename = os.path.basename(deb_file)
    sudo('dpkg -i {0}'.format(deb_filename), shell=False)


LOCAL_CONFIG = """
[__main__]
includes = config/staging.cfg
    config/dailies.cfg
"""


def update_configs(config_branch, config_dir='django_project'):
    """Backup current config files and pull in the latest from a bzr branch.

    If it exists, local.cfg is left unmodified.
    """
    timestamp = datetime.utcnow().isoformat()
    backup_dir = "%s.backup.%s" % (config_dir, timestamp)
    if exists(config_dir):
        run("mv %s %s" % (config_dir, backup_dir))
    put_bzr_branch(config_branch, config_dir)
    if exists(backup_dir + '/local.cfg'):
        run("cp %s/local.cfg %s" % (backup_dir, config_dir))
    else:
        # Create a local.cfg on the server from our template.
        local_file, local_path = tempfile.mkstemp()
        local_file.write(LOCAL_CONFIG)
        local_file.close()
        put(local_path, "%s/local.cfg" % config_dir)
        os.remove(local_path)


def put_bzr_branch(url, name):
    """
    Copy bazaar branch from launchpad to remote host

    This have to be done with local machine as proxy, because remote machine
    cannot have access to private projects on launchpad. We don't want to put
    there our private ssh keys.

    :param url: bazaar url to the branch, it can start with lp:
    :type  url: str

    :param name: name of the directory in which branch will be located remotely
    :type  name: str

    """
    local("rm -rf /tmp/%s" % name)

    local("bzr branch %s /tmp/%s" % (url, name))
    local("cd /tmp && tar cjf %s.tar.bz2 %s" % (name, name))
    put("/tmp/%s.tar.bz2" % name, '~/')

    local("rm -rf /tmp/%s.tar.bz2 /tmp/%s" % (name, name))

    run("tar xjf %s.tar.bz2" % name)


def build_and_install_deb(config_branch, build_dir=None, distroseries='lucid'):
    delete_dir_after_build = False
    if build_dir is None:
        build_dir = tempfile.mkdtemp()
        delete_dir_after_build = True

    dsc_file = build_source_package(build_dir, distroseries=distroseries)
    binary_file = build_binary_package(build_dir, dsc_file)
    install_binary_package(binary_file)
    migrate_database()
    update_configs(config_branch)
    preflight_check()
    restart_services()
    print("Package installed and services restarted.")

    if delete_dir_after_build:
        shutil.rmtree(build_dir, ignore_errors=True)
