# -*- coding: utf-8 -*-
#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Piston handlers for the reviews-and-ratings API."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'FlagReviewHandler',
    'ReviewsStatsHandler',
    'ServerStatusHandler',
    'ShowReviewsHandler',
    'SubmitReviewHandler',
    'ShowUsefulnessHandler',
]

import urllib
from datetime import datetime, timedelta

from django.conf import settings
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseNotFound,
    )
from django.shortcuts import get_object_or_404
from django.utils import simplejson
from piston.handler import BaseHandler

from reviewsapp.forms import ReviewForm
from reviewsapp.models import (
    Repository,
    Review,
    RNRSettings,
    SoftwareItem,
    SoftwareItemInOrigin,
    SoftwareItemInRepository,
    Usefulness,
    )
from reviewsapp.utilities import invalidate_paginated_reviews_for


class ServerStatusHandler(BaseHandler):
    allowed_methods = ('GET',)

    def read(self, request):
        return "ok"


class ReviewsStatsHandler(BaseHandler):
    allowed_methods = ('GET',)

    def read(self, request, origin='any', distroseries='any', days=None):
        # We do not allow requests with GET parameters which avoid
        # django's caching system (risk of DoS attack).
        if request.GET:
            return HttpResponseBadRequest(
                "This api call does not take GET parameters.")

        # This explicit extra selects seem necessary to ensure that the
        # extracted values below can extract just 'package_name' rather
        # than 'softwareitem__package_name' etc.
        # XXX 2011-03-18 michaeln bug=737649
        # We include a blank app_name until the USC client no longer
        # expects the app_name attribute in the stats.
        extra_options = {
            'select': dict(
                package_name='"reviewsapp_softwareitem"."package_name"',
                app_name="''"),
            'tables': ['reviewsapp_softwareitem'],
            'where': ['"softwareitem_id" = "reviewsapp_softwareitem"."id"'],
            }

        if origin == distroseries == 'any':
            items = SoftwareItem.objects.all().extra(
                select=dict(app_name="''"))
        elif distroseries == 'any':
            items = SoftwareItemInOrigin.objects.filter(origin=origin).extra(
                **extra_options)
        elif origin == 'any':
            # Note: If we eventually have the same package in multiple
            # origins (ie. when we support PPAs generally) we will need
            # to add the origin/distroseries attributes to the json
            # result. Also, we are returning the stats for the item in a
            # repository (ie. specific origin/distroseries), not the
            # stats for an item in a distroseries across multiple
            # repositories.
            items = SoftwareItemInRepository.objects.filter(
                repository__distroseries=distroseries).extra(**extra_options)
        else:
            try:
                repository = Repository.objects.get(
                    origin=origin, distroseries=distroseries)
            except Repository.DoesNotExist:
                return []

            items = SoftwareItemInRepository.objects.filter(
                repository=repository).extra(**extra_options)

        if days is not None:
            date_from = datetime.utcnow() - timedelta(int(days))
            items = items.filter(date_ratings_updated__gt=date_from)

        return items.values('package_name', 'app_name', 'ratings_total',
            'ratings_average')


class RetrieveReviewHandler(BaseHandler):
    allowed_methods = ('GET',)

    def read(self, request, review_id):
        try:
            review = Review.objects.get(id=review_id)
            return review
        except Review.DoesNotExist:
            return HttpResponseNotFound('No matching review found')

class ShowUsefulnessHandler(BaseHandler):
    model = Usefulness
    fields = (
        'review_id',
        'username',
        'useful',
    )
    allowed_methods = ('GET',)

    def read(self, request):
        username = request.GET.get('username', None)
        review_id = request.GET.get('review_id', None)

        if not username and not review_id:
            return HttpResponseBadRequest('No valid parameters passed.'
               + ' Please pass at least username or review_id')

        if username:
            response = Usefulness.objects.filter(user__username=username)
            if review_id:
                response = response.filter(
                    review__id=review_id)
        else:
            response = Usefulness.objects.filter(
                review__id=review_id)

        return response

class ShowReviewsHandler(BaseHandler):
    model = Review
    fields = (
        'id',
        'package_name',
        'app_name',
        'origin',
        'distroseries',
        'language',
        'date_created',
        'rating',
        'reviewer_username',
        'reviewer_displayname',
        'summary',
        'review_text',
        'hide',
        'version',
        'usefulness_total',
        'usefulness_favorable',
    )
    allowed_methods = ('GET',)

    def read(self, request, language, origin, distroseries, version,
             package_name, page='1'):
        # Django passes all params as unicode strings, but we need to
        # end up with an unquoted bytestring. That is, for Django,
        # SoftwareItem.objects.get(app_name=u'Déjà Dup') is equivalent
        # to
        # SoftwareItem.objects.get(app_name='D\xc3\x83\xc2\xa9j\xc3\x83\xc2\xa0
        # Dup') but not:
        # SoftwareItem.objects.get(app_name=u'D\xc3\x83\xc2\xa9j\xc3\x83\xc2\xa0
        # Dup')
        software_item = urllib.unquote_plus(
            urllib.unquote_plus(str(package_name)))
        package_name, sep, appname = software_item.partition(';')
        batch_size = settings.API_REVIEWS_PAGINATION
        batch_start = batch_size * (int(page) - 1)
        batch_end = batch_start + batch_size
        kwargs = {
            'language': language,
            'repository__distroseries': distroseries,
            'repository__origin': origin,
            'version': version,
        }
        kwargs = dict(item for item in kwargs.items() if item[1] != 'any')
        return Review.objects.filter(
            softwareitem__package_name=package_name, hide=False,
            **kwargs).order_by(
                '-usefulness_percentage')[batch_start:batch_end]


class SubmitReviewHandler(BaseHandler):
    allowed_methods = ('POST',)

    def create(self, request):
        review = Review(reviewer=request.user)
        form = ReviewForm(request.data, instance=review)
        if not form.is_valid():
            errors = dict((key, [unicode(v) for v in values])
                for (key, values) in form.errors.items())
            response = simplejson.dumps({'errors': errors})
            return HttpResponseBadRequest(response)

        review = form.save()

        # Automatically flag when appropriate.
        if RNRSettings.should_moderate_text(review.summary, review.review_text):
            mod_sum, mod_desc = RNRSettings.get_moderation_summary_and_desc()
            review.flag(None, mod_sum, mod_desc)

        # Update stats on the software item
        review.softwareitem_in_repository.update_stats()

        # Invalidate the cache for this package.
        invalidate_paginated_reviews_for(review)

        return review


class FlagReviewHandler(BaseHandler):
    allowed_methods = ('POST',)

    def create(self, request, review_id):
        try:
            # XXX achuni 2010-11-25: We should restrict reason/text further!
            #      Impose a max length, and probably don't allow \n if we're
            #      counting on being able to join with that.
            reason = request.POST['reason']
            request_text = request.POST['text']
        except KeyError as error:
            return HttpResponseBadRequest("Missing required argument '%s'"
                % str(error))
        # Find the referenced review.
        try:
            review = Review.objects.get(id=review_id)
        except Review.DoesNotExist:
            return HttpResponseNotFound('No matching review found')
        # Create a moderation request for this review, and invalidate
        # the cache if the review is hidden as a result.
        review_was_hidden = review.hide
        flag = review.flag(request.user, summary=reason,
                           description=request_text)
        if review.hide and not review_was_hidden:
            invalidate_paginated_reviews_for(review)
        return flag


class SubmitUsefulnessHandler(BaseHandler):
    allowed_methods = ('POST',)

    def create(self, request, review_id):
        """Cast a vote on whether or not the review was useful. """
        review = get_object_or_404(Review, id=review_id)
        usefulness, created = Usefulness.objects.get_or_create(
            review=review, user=request.user)
        old_useful = usefulness.useful
        useful = request.POST.get('useful', 'True')
        usefulness.useful = (useful != 'False')
        usefulness.save()
        total = Usefulness.objects.filter(review=review).count()
        favorable = Usefulness.objects.filter(review=review,
                                              useful=True).count()
        percentage = favorable / float(total)
        review.usefulness_total = total
        review.usefulness_favorable = favorable
        review.usefulness_percentage = int(percentage * 100)  # for sorting
        review.save()

        # note that we need the (double) quotes around the reply as this
        # is interpreted by piston-mini-client as a json reply
        if created:
            return HttpResponse('"Created"', status=201)
        elif  old_useful != usefulness.useful:
            return HttpResponse('"Updated"', status=200)
        else:
            return HttpResponse('"Not modified"', status=200)



