#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.conf.urls.defaults import url, patterns
from django.views.decorators.cache import (
    cache_control,
    never_cache,
    )

from piston.resource import Resource
from reviewsapp.api.handlers import (
    FlagReviewHandler,
    ReviewsStatsHandler,
    ServerStatusHandler,
    ShowReviewsHandler,
    SubmitReviewHandler,
    SubmitUsefulnessHandler,
    RetrieveReviewHandler,
    ShowUsefulnessHandler
)
from reviewsapp.auth import SSOOAuthAuthentication
from reviewsapp.api.decorators import dont_vary

auth = SSOOAuthAuthentication(realm="Ratings and Reviews")

# The server status resource should never be cached.
server_status_resource = Resource(handler=ServerStatusHandler)
server_status_resource.__name__ = "Resource"
server_status_resource = never_cache(server_status_resource)

# Although the review stats resource will be cached by the cache
# middleware, we update the cache-control so that upstream caches will
# keep it around for much longer (CACHE_REVIEW_STATS_SECONDS).
reviews_stats_resource = Resource(handler=ReviewsStatsHandler)
cached_resource = cache_control(max_age=settings.CACHE_REVIEW_STATS_SECONDS)
reviews_stats_resource.__name__ = "Resource"
reviews_stats_resource = dont_vary(cached_resource(reviews_stats_resource))

# The show reviews resource is cached.
show_reviews_resource = Resource(handler=ShowReviewsHandler)
show_reviews_resource.__name__ = "Resource"
show_reviews_resource = dont_vary(show_reviews_resource)

retrieve_review_resource = Resource(handler=RetrieveReviewHandler)
retrieve_review_resource.__name__ = "Resource"
retrieve_review_resource = dont_vary(retrieve_review_resource)

# The show usefulness resource is cached.
show_usefulness_resource = Resource(handler=ShowUsefulnessHandler)
show_usefulness_resource.__name__ = "Resource"
show_usefulness_resource = dont_vary(show_usefulness_resource)


# The following POST handlers have POST data and will not be cached.
submit_review_resource = Resource(handler=SubmitReviewHandler,
    authentication=auth)
flag_review_resource = Resource(handler=FlagReviewHandler,
    authentication=auth)
submit_usefulness_resource = Resource(handler=SubmitUsefulnessHandler,
    authentication=auth)

urlpatterns = patterns('',
    # get status of the service (usually just "ok", might be "read-only")
    # to ensure we inform the user when he wants to write a review or
    # send a moderation request
    #  GET /1.0/server-status/
    url(r'^1.0/server-status/$', server_status_resource,
        name='rnr-server-status'),

    # gets the ratings for a given distroseries
    #  GET /1.0/review-stats
    url(r'^1.0/review-stats/(?P<origin>[-\w]+)/(?P<distroseries>\w+)/$',
        reviews_stats_resource, name='rnr-api-reviews-stats'),
    url(r'^1.0/review-stats/(?P<origin>[-\w]+)/(?P<distroseries>\w+)/'
        r'updates-last-(?P<days>[1,3,7])-days/$',
        reviews_stats_resource, name='rnr-api-reviews-stats'),

    # The following urls are kept for backwards compatibility with old
    # versions of rnr-client (rnrclient <= r33). Once the bulk of
    # natty clients have been updated, we can remove these two and the
    # associated tests.
    url(r'^1.0/review-stats/$', reviews_stats_resource,
        name='rnr-api-reviews-stats'),
    url(r'^1.0/review-stats/updates-last-(?P<days>[1,3,7])-days/$',
        reviews_stats_resource, name='rnr-api-reviews-stats'),

    # gets the reviews (and ratings) for a given pkg/app
    #  GET /1.0/en/ubuntu/+binary/lucid/1:2.32.0-0ubuntu1/gnome-games;sane/page/1/
    #  GET /1.0/en/ubuntu/+binary/lucid/1:2.32.0-0ubuntu1/2vcard/
    url(r'^1.0/reviews/filter/(?P<language>\w+)/(?P<origin>[-\w]+)/'
        r'(?P<distroseries>\w+)/(?P<version>[-\w+.:~]+)/'
        r'(?P<package_name>[\w%+.;-]+)/$',
        show_reviews_resource, name='rnr-api-reviews-for-package'),
    url(r'^1.0/reviews/filter/(?P<language>\w+)/(?P<origin>[-\w]+)/'
        r'(?P<distroseries>\w+)/(?P<version>[-\w+.:~]+)/'
        r'(?P<package_name>[\w%+.;-]+)/page/(?P<page>\d+)/$',
        show_reviews_resource, name='rnr-api-reviews-for-package'),

    # vote on the usefulness of a review
    #  POST /1.0/reviews/405/usefulness/False/
    url(r'^1.0/reviews/(?P<review_id>\d+)/recommendations/$',
        submit_usefulness_resource, name='rnr-api-recommendations'),

    # get the filtered usefulness records using parameters for filtering
    #  GET /1.0/reviews/usefulness/?username=aaronp
    #  GET /1.0/reviews/usefulness/?review_id=100
    #  GET /1.0/reviews/usefulness/?username=aaronp&review_id=100
    url(r'^1.0/usefulness/$',
        show_usefulness_resource, name='rnr-api-usefulness-filter'),

    # gets a single review for a given review id
    # GET /1.0/reviews/100/
    url(r'^1.0/reviews/(?P<review_id>\d+)/$',
        retrieve_review_resource, name='rnr-api-review-for-review-id'),

    # submits a new review for a given pkg/app
    #  POST /1.0/reviews/
    url(r'^1.0/reviews/$', submit_review_resource, name='rnr-api-reviews'),

    # flag inappropriate reviews
    #  POST /1.0/reviews/1/+report-review/
    url(r'^1.0/reviews/(?P<review_id>\d+)/flags/$',
        flag_review_resource),
)
