#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SSO/OAuth authenticator for Piston."""


from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'SSOOAuthAuthentication',
    ]

from datetime import datetime, timedelta
from django.conf import settings
from django.contrib.auth.models import User
from django.http import HttpResponse
from oauth import oauth
from piston.authentication import OAuthAuthentication, oauth_datastore
from piston.oauth import OAuthError

from reviewsapp.models import Consumer, Token
from reviewsapp.utilities import (
    WebServices,
    full_claimed_id,
)

TOKEN_CACHE_EXPIRY = timedelta(hours=
    getattr(settings, 'TOKEN_CACHE_EXPIRY_HOURS', 4))

class SSOOAuthAuthentication(OAuthAuthentication):
    """ This class is a Piston Authentication class.

    See http://bitbucket.org/jespern/django-piston/wiki/Documentation for
    more info.
    """
    def __init__(self, realm='API'):
        self.realm = realm

    def challenge(self):
        """ We define our own challenge so that we can provide our own realm
        and return a simple error message instead of rendering a template.
        """
        resp = HttpResponse("Authorization Required")
        resp['WWW-Authenticate'] = 'OAuth realm="%s"' % self.realm
        resp.status_code = 401
        return resp

    def is_authenticated(self, request):
        """ We override is_authenticated so we can prefetch creds from SSO
        before the oauth mechanism asks us for them
        """
        headers = {
            'Authorization' : request.META.get('HTTP_AUTHORIZATION', '')
        }
        orequest = oauth.OAuthRequest.from_request(
            request.method, request.build_absolute_uri(), headers=headers,
            query_string=request.META['QUERY_STRING'])
        if (orequest is None or
            not 'oauth_token' in orequest.parameters or
            not 'oauth_consumer_key' in orequest.parameters):
            return False

        self.prefetch_oauth_consumer(orequest)
        return super(SSOOAuthAuthentication, self).is_authenticated(request)

    def prefetch_oauth_consumer(self, request):
        """Check if we have a consumer for the current creds, and validate
           the token with SSO if we don't.
        """
        web_services = WebServices()
        oauthtoken = request.get_parameter('oauth_token')
        consumer_key = request.get_parameter('oauth_consumer_key')
        tokens = Token.objects.filter(token=oauthtoken,
            consumer__key=consumer_key)
        if len(tokens) == 0 or (tokens[0].updated_at <
            datetime.now() - TOKEN_CACHE_EXPIRY):
            pieces = web_services.get_data_for_account(token=oauthtoken,
                openid_identifier=consumer_key, signature=request.get_parameter('oauth_signature'))
            if not pieces:
                return
            Consumer.objects.filter(key=consumer_key).exclude(
                secret=pieces['consumer_secret']).delete()
            claimed_id = full_claimed_id(consumer_key)
            try:
                user = User.objects.get(
                    useropenid__claimed_id=claimed_id)
            except User.DoesNotExist:
                if (not pieces.get('preferred_email') or
                    not pieces.get('username')):
                    return
                user = User.objects.create_user(pieces['username'],
                    pieces['preferred_email'], password=None)
                user.useropenid_set.create(claimed_id=claimed_id)

            displayname = pieces['displayname']
            if displayname != user.get_full_name():
                user.first_name, sep, user.last_name = displayname.rpartition(
                    ' ')
                user.save()

            consumer, created = Consumer.objects.get_or_create(user=user,
                key=consumer_key, secret=pieces['consumer_secret'])
            token, created = Token.objects.get_or_create(consumer=consumer,
                token=pieces['token'], token_secret=pieces['token_secret'],
                name=pieces['name'])
            if not created:
                # Update updated_at
                token.save()

    @staticmethod
    def validate_token(request, check_timestamp=True, check_nonce=True):
        oauth_server, oauth_request = initialize_server_request(request)
        if oauth_server is None:
            raise OAuthError('initialize_server_request returned None')
        return oauth_server.verify_request(oauth_request)

def initialize_server_request(request):
    """
    Shortcut for initialization.
    """
    headers = {
        'Authorization' : request.META.get('HTTP_AUTHORIZATION', '')
    }
    oauth_request = oauth.OAuthRequest.from_request(
        request.method, request.build_absolute_uri(), headers=headers,
        query_string=request.META['QUERY_STRING'])

    if oauth_request:
        oauth_server = oauth.OAuthServer(oauth_datastore(oauth_request))
        oauth_server.add_signature_method(oauth.OAuthSignatureMethod_PLAINTEXT())
        oauth_server.add_signature_method(oauth.OAuthSignatureMethod_HMAC_SHA1())
    else:
        oauth_server = None

    return oauth_server, oauth_request
