# Copyright 2010 Canonical Ltd.
#
# This file is part of Software Center Ratings and Reviews.
#
# Software Center Ratings and Reviews is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Software Center Ratings and Reviews is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Software Center Ratings and Reviews.  If not, see
# <http://www.gnu.org/licenses/>.

"""Forms for validation of review moderations."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'ReviewModerationFilterForm',
    'ReviewModerationModerateForm',
    'ReviewForm',
    ]

from django import forms
from django.conf import settings
from django.db.models import Count
from django.utils.translation import ugettext_lazy as _

from reviewsapp.models import (
    Architecture,
    Repository,
    Review,
    ReviewModeration,
    SoftwareItem,
    )
from reviewsapp.utilities import WebServices
from reviewsapp.widgets import MultipleSubmitButton


class ReviewModerationModerateForm(forms.Form):
    """A simple form that can display itself."""
    moderation_text = forms.CharField(required=False)

    status = forms.TypedChoiceField(
        widget=MultipleSubmitButton,
        choices=ReviewModeration.MODERATION_CHOICES,
        coerce=int)


class ReviewForm(forms.ModelForm):

    # The following extra fields are used to calculate their model
    # equivalents.
    package_name = forms.CharField()
    app_name = forms.CharField(required=False)
    version = forms.CharField(max_length=100)
    origin = forms.CharField()
    distroseries = forms.CharField()
    arch_tag = forms.CharField(max_length=16)

    class Meta:
        model = Review
        fields = ('package_name', 'app_name', 'summary', 'review_text',
                  'rating', 'language')

    def save(self, commit=True):
        """Save the extra model data obtained from fields above."""
        review = super(ReviewForm, self).save(commit=False)

        review.softwareitem = self.cleaned_data['software_item']
        review.version = self.cleaned_data['version']
        review.repository = self.cleaned_data['repository']
        review.architecture = self.cleaned_data['architecture']
        if commit:
            review.save()
        return review

    def clean(self):
        """Add the software_item to the cleaned data, creating if necessary.
        """
        # If the supplied data is not valid, no point in continuing.
        if not self.is_valid():
            return self.cleaned_data

        self._validate_software_item_and_repository()

        if not settings.ALLOW_MULTIPLE_REVIEWS_FOR_TESTING:
            # Ensure a user cannot submit a second review for the same app
            previous_reviews = Review.objects.filter(
                reviewer=self.instance.reviewer,
                softwareitem=self.cleaned_data['software_item'],
                repository=self.cleaned_data['repository'])
            if previous_reviews:
                raise forms.ValidationError(
                    "A user cannot create multiple reviews for an app.")

        return self.cleaned_data

    def clean_arch_tag(self):
        """Validate the architecture tag and replace with the model."""
        arch_tag = self.cleaned_data['arch_tag']

        try:
            architecture = Architecture.objects.get(tag=arch_tag)
        except Architecture.DoesNotExist:
            raise forms.ValidationError(
                "The specified architecture '{0}' does not exist.".format(
                    arch_tag))

        self.cleaned_data['architecture'] = architecture
        return arch_tag

    def _validate_software_item_and_repository(self):
        """Add the software item and repository to the cleaned data.

        If they do not exist in the db, we check the validity and
        create them if they are valid.
        """
        package_name = self.cleaned_data['package_name']
        app_name = self.cleaned_data['app_name']
        origin = self.cleaned_data['origin']
        distroseries = self.cleaned_data['distroseries']
        arch_tag=self.cleaned_data['arch_tag']

        require_repository_check = False
        repository = None
        # First we grab the repository - if it exists.
        try:
            repository = Repository.objects.get(
                origin=origin, distroseries=distroseries)
        except Repository.DoesNotExist:
            require_repository_check = True

        if repository is not None:
            # Check whether there are any existing reviews for this
            # software in the repository with the given arch.
            review_count = Review.objects.filter(
                softwareitem__package_name=package_name,
                repository=repository,
                architecture__tag=arch_tag).count()
            if review_count == 0:
                # This is the first review for this software_item in the
                # repository with the given arch.
                require_repository_check = True

        if require_repository_check:
            ws = WebServices()
            valid_in_lp = ws.lp_verify_packagename_in_repository(
                package_name, origin, distroseries, arch_tag=arch_tag)

            if not valid_in_lp:
                valid_in_sca = ws.sca_verify_packagename_in_repository(
                    package_name, origin, distroseries, arch_tag=arch_tag)
                if not valid_in_sca:
                    raise forms.ValidationError(
                        ': package {0} not in {1} {2} for {2}'.format(
                            package_name, origin, distroseries, arch_tag))

        # Now we know the given data corresponds to a published binary
        # on launchpad, so we can populate the cleaned data, creating
        # the repository and software_item records if necessary.
        repository = repository or Repository.objects.create(
            origin=origin, distroseries=distroseries)
        self.cleaned_data['repository'] = repository

        software_item, created = SoftwareItem.objects.get_or_create(
            package_name=package_name)
        self.cleaned_data['software_item'] = software_item


class ReviewModerationFilterForm(forms.Form):
    """Filtering options for review moderation list."""

    language = forms.ChoiceField(required=False,
        choices=(('', _('All languages')),))
    distro_series = forms.CharField(
        required=False, widget=forms.HiddenInput())
    status = forms.TypedChoiceField(
        widget=forms.HiddenInput(),
        choices=ReviewModeration.STATUS_CHOICES,
        empty_value=ReviewModeration.PENDING_STATUS,
        coerce=int, required=False)

    def get_language_choices(self):
        # Dynamically set add the current languages (including the
        # number of moderations for each).
        moderations = ReviewModeration.objects.all()

        # If we were initialised with a status, then filter the language
        # choices accordingly.
        try:
            status = self.fields['status'].clean(self.data.get('status'))
        except forms.ValidationError:
            status = ReviewModeration.PENDING_STATUS

        moderations = moderations.filter(status=status)

        # Grab the languages and counts for each, and format for the
        # field choices. Ie. ('de', 'de (2)', 'en', 'en (2)')
        languages = moderations.values('review__language').annotate(
            Count('review__language')).order_by('review__language')
        language_choices = [(
            lang_count['review__language'],
            lang_count['review__language'] + " ({0})".format(
                lang_count['review__language__count']),
            ) for lang_count in languages]
        language_choices.insert(0, ('', _('All languages')))
        return language_choices

    def __init__(self, *args, **kwargs):
        """Set the language choices dynamically."""
        super(ReviewModerationFilterForm, self).__init__(*args, **kwargs)

        self.fields['language'].choices = self.get_language_choices()
