from datetime import datetime
from optparse import make_option
from os import urandom
from random import choice, randint, random
from string import lowercase
from time import time

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand
from django.db import connection

from reviewsapp.models import (
    Repository,
    Architecture,
)

lorem_words = """Lorem ipsum dolor sit amet, consectetur adipiscing elit.
Mauris eget sapien vitae orci tincidunt rutrum. Nulla sagittis elit quis ipsum
mattis cursus. Suspendisse id elit est, et ornare nulla. In at magna a lacus
tincidunt rutrum non ut lectus. Aliquam erat volutpat. Vestibulum facilisis
tempus erat, et auctor odio interdum eget. Phasellus dapibus magna at sapien
fringilla sollicitudin. Donec nec lorem metus. Phasellus erat justo, luctus
in auctor sit amet, laoreet sed enim. Aenean eget ligula libero, at hendrerit
magna. Vivamus lobortis malesuada mauris, ac suscipit risus cursus quis. Duis
sed orci at quam mollis porta ac eget risus. Praesent facilisis est eget leo
placerat imperdiet. Sed dictum aliquet bibendum. Pellentesque quis ipsum eu
risus tempor blandit. Nunc scelerisque arcu tristique turpis vulputate ac
lobortis risus hendrerit. Nam arcu lorem, imperdiet a posuere quis, tempor
ac lorem.""".split()

class LenGenerator(object):
    """An generator that knows its own length"""
    def __init__(self, generator, length):
        self.length = length
        self.generator = generator
    def __iter__(self):
        return self
    def next(self):
        return self.generator.next()
    def __len__(self):
        return self.length

class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option('--softwareitems', dest='softwareitems', default=1000,
            help='Number of random software items to create.'),
        make_option('--reviews', dest='reviews', default=100000,
            help='Number of random reviews to create.'),
        make_option('--sessions', dest='sessions', default=100000,
            help='Number of random sessions to create.'),
        make_option('--users', dest='users', default=100000,
            help='Number of random users to create.'),
    )
    help = "Populates the Database with a steaming pile of random data."

    def populate_table(self, tablename, rowgenerator, nrows, numeric_id=True):
        cursor = connection.cursor()
        if self.verbosity >= 2:
            print "Populating %s..." % tablename
        if numeric_id:
            cursor.execute("select max(id) from %s" % tablename)
            max_val = cursor.fetchone()[0]
            if max_val is None:
                first_val = 0
            else:
                first_val = max_val + 1
        else:
            first_val = 0
        self.ranges[tablename] = (first_val, first_val + nrows - 1)
        ncols = len(rowgenerator(0))
        placeholders = ', '.join([self.placeholder] * ncols)
        rows = LenGenerator(
            (rowgenerator(x) for x in xrange(first_val, first_val + nrows)),
            nrows
        )
        query = "insert into %s values (%s)" % (tablename, placeholders)
        cursor.executemany(query, rows)
        connection.connection.commit()


    def handle(self, *args, **options):
        nsessions = int(options['sessions'])
        nusers = int(options['users'])
        nsoftwareitems = int(options['softwareitems'])
        nreviews = int(options['reviews'])
        self.verbosity = int(options['verbosity'])
        self.translation_table = (lowercase * 10)[:256]
        self.ranges = {}
        if settings.DATABASE_ENGINE == 'postgresql_psycopg2':
            self.placeholder = '%s'
            self.booleans = ['f', 't']
        elif settings.DATABASE_ENGINE == 'sqlite3':
            self.placeholder = '?'
            self.booleans = [0, 1]
        else:
            raise ImproperlyConfigured("Only postgresql_psycopg2 and sqlite3"
                " backends are supported by the populate command.")

        lucid, created = Repository.objects.get_or_create(origin='ubuntu',
            distroseries='lucid')
        maverick, created = Repository.objects.get_or_create(origin='ubuntu',
            distroseries='maverick')
        self.ranges['reviewsapp_repository'] = (lucid.id, maverick.id)

        i386, created = Architecture.objects.get_or_create(tag='i386')
        amd64, created = Architecture.objects.get_or_create(tag='amd64')
        self.ranges['reviewsapp_architecture'] = (i386.id, amd64.id)

        if nsessions > 0:
            self.populate_table('django_session', self.gen_session, nsessions,
                numeric_id=False)
        if nusers > 0:
            self.populate_table('auth_user', self.gen_user, nusers)
        if nsoftwareitems > 0:
            self.populate_table('reviewsapp_softwareitem',
                self.gen_softwareitem, nsoftwareitems)
        if nreviews > 0:
            self.populate_table('reviewsapp_review', self.gen_review,
                nreviews)

    def random_string(self, size):
        msg = urandom(size)
        return msg.translate(self.translation_table)

    def random_email(self):
        base = self.random_string(20)
        return '%s@%s.%s' % (base[:9], base[9:17], base[17:])

    def random_comment(self, wordcount=10):
        """A naive attempt at generating something that looks like text"""
        comment_words = [choice(lorem_words) for i in range(wordcount)]
        return ' '.join(comment_words).capitalize() + '.'

    def random_date(self):
        year = randint(2009, 2011)
        month = randint(1, 12)
        day = randint(1, 28)
        return str(datetime(year, month, day))

    def gen_session(self, id):
        """ Returns a sequence representing a Django session.

        Currently (session_key, session_data, expire_date).
        """
        idstr = self.random_string(size=12) + str(id)
        return[idstr, self.random_string(size=200), self.random_date()]

    def gen_user(self, id):
        """ Returns a sequence representing a User.

        Currently (id, username, first_name, last_name, email, password,
            is_staff, is_active, is_superuser, last_login, date_joined)
        """
        first_name = self.random_string(8).capitalize()
        last_name = self.random_string(8).capitalize()
        username = first_name + last_name
        email = self.random_email()
        date_joined = self.random_date()
        false = self.booleans[0]
        true = self.booleans[1]
        return [id, username, first_name, last_name, email, '!', false, true,
            false, date_joined, date_joined]

    def gen_softwareitem(self, id):
        """ Returns a sequence representing a SoftwareItem.

        Currently (id, ratings_total, ratings_average,
            date_ratings_updated, package_name)
        """
        package_name = self.random_string(8)
        ratings_updated = datetime.now()
        # XXX: Link these to the reviews we're generating!
        ratings_total = randint(0, 100)
        ratings_average = random() * 5
        return [id, ratings_total, ratings_average, ratings_updated,
                package_name]

    def gen_review(self, id):
        """ Returns a sequence representing a Review.

        Currently (id, softwareitem_id, version, architecture_id,
            repository_id, reviewer_id, date_created, rating, summary,
            review_text, language, hide, usefulness_total,
            usefulness_favorable, usefulness_percentage, app_name)
        """
        softwareitem_id = randint(*self.ranges['reviewsapp_softwareitem'])
        version = '%.1f' % random()
        arch_id = choice(self.ranges['reviewsapp_architecture'])
        repo_id = choice(self.ranges['reviewsapp_repository'])
        reviewer_id = randint(*self.ranges['auth_user'])
        date_created = self.random_date()
        rating = randint(1, 5)
        summary = self.random_comment(wordcount=5)
        review_text = self.random_comment(wordcount=20)
        false = self.booleans[0]
        return [id, softwareitem_id, version, arch_id, repo_id, reviewer_id,
            date_created, rating, summary, review_text, 'en', false, 0, 0, 0, '']

