# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging
import re
import sys
import traceback

from django.conf import settings
from django.db import connection
from django.views import debug
from django.views.debug import ExceptionReporter

default_hidden_settings = \
    'BIBIT|SECRET|PASSWORD|PROFANITIES_LIST|PRIVATE|secret|password|private'
hidden_settings = getattr(settings, 'HIDDEN_SETTINGS', default_hidden_settings)
debug.HIDDEN_SETTINGS = re.compile(hidden_settings)


class SanitizedExceptionReporter(ExceptionReporter):
    def __init__(self, request, exc_type, exc_value, tb):
        request = self.sanitize_request(request)
        ExceptionReporter.__init__(self, request,
                                         exc_type, exc_value, tb)

    def get_traceback_frames(self):
        # This method is taken from Django's views.debug.ExceptionReporter.
        # Please see the license file in the third-party/django directory.
        frames = []
        tb = self.tb
        while tb is not None:
            # support for __traceback_hide__ which is used by a few libraries
            # to hide internal frames.
            if tb.tb_frame.f_locals.get('__traceback_hide__'):
                tb = tb.tb_next
                continue
            filename = tb.tb_frame.f_code.co_filename
            function = tb.tb_frame.f_code.co_name
            lineno = tb.tb_lineno - 1
            loader = tb.tb_frame.f_globals.get('__loader__')
            module_name = tb.tb_frame.f_globals.get('__name__')
            pre_context_lineno, pre_context, context_line, post_context = \
                    self._get_lines_from_file(filename, lineno, 7, loader,
                                              module_name)
            if pre_context_lineno is not None:
                frames.append({
                    'tb': tb,
                    'filename': filename,
                    'function': function,
                    'lineno': lineno + 1,
                    # The only diff in this method: call self.sanitize_vars()
                    'vars': self.sanitize_vars(tb.tb_frame.f_locals.items()),
                    'id': id(tb),
                    'pre_context': pre_context,
                    'context_line': context_line,
                    'post_context': post_context,
                    'pre_context_lineno': pre_context_lineno + 1,
                })
            tb = tb.tb_next

        if not frames:
            frames = [{
                'filename': '&lt;unknown&gt;',
                'function': '?',
                'lineno': '?',
                'context_line': '???',
            }]

        return frames

    # These two methods are new -- Sanitize the request's POST and GET dicts,
    # and local variables in frames, not only Django's settings.
    def sanitize_vars(self, items):
        sanitized = []
        for item in items:
            if debug.HIDDEN_SETTINGS.search(item[0]):
                sanitized.append((item[0], '********'))
            else:
                sanitized.append(item)
        return sanitized

    def sanitize_request(self, request):
        """Remove sensitive from the request before it is displayed"""
        dup_post = request.POST.copy()
        for key in request.POST.iterkeys():
            if debug.HIDDEN_SETTINGS.search(key):
                dup_post[key] = '********'
        request.POST = dup_post
        dup_get = request.GET.copy()
        for key in request.GET.iterkeys():
            if debug.HIDDEN_SETTINGS.search(key):
                dup_get[key] = '********'
        request.GET = dup_get
        return request


debug.ExceptionReporter = SanitizedExceptionReporter


class LogExceptionMiddleware(object):
    def process_exception(self, request, exception):
        """Log the traceback, so that it can be kept in the oops file"""
        reporter = SanitizedExceptionReporter(request, *sys.exc_info())
        template_debug = settings.TEMPLATE_DEBUG
        settings.TEMPLATE_DEBUG = True
        try:
            logging.warn(traceback.format_exc())
            for query in connection.queries:
                logging.warn("time: %(time)s sql: %(sql)s" % query)
        finally:
            settings.TEMPLATE_DEBUG = template_debug
