#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time

from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed
from linux_metrics import (
    cpu_stat,
    disk_stat,
    mem_stat,
    net_stat,
)


class StatisticsLoggingMiddleware(object):
    """When enabled this middleware will log vital stats like CPU load, RAM
    usage and metwork activity.

    To disable this middleware, set settings.ENABLE_STATISTICS_LOGGING to
    False

    To configure the stats that are logged change the settings.STATS_TO_LOG
    setting as a sum of LOG_CPU, LOG_MEMORY, LOG_DISK and LOG_NETWORK and
    including only those that are to be logged.
    """
    last_run = time.time()
    LOG_CPU = 'cpu'
    LOG_MEMORY = 'ram'
    LOG_DISK = 'disk'
    LOG_NETWORK = 'net'

    def __init__(self):
        if not settings.ENABLE_STATISTICS_LOGGING:
            raise MiddlewareNotUsed()
        else:
            self.logger = logging.getLogger('rnr-statistics-logging')
            self.interfaces = settings.NETWORK_INTERFACES
            self.disks = settings.DISKS
            self.seconds_between_logging = settings.SECONDS_BETWEEN_LOGGING
            self.stats_to_log = settings.STATS_TO_LOG

    def process_response(self, requset, response):
        now = time.time()
        diff = now - StatisticsLoggingMiddleware.last_run
        if diff > self.seconds_between_logging:
            self.log_stats()
            StatisticsLoggingMiddleware.last_run = now
        return response

    def log_stats(self):
        try:
            stats = []
            if self.LOG_CPU in self.stats_to_log:
                stats.append('CPU: percents & times ')
                stats.append(str(cpu_stat.cpu_percents()))
                stats.append(str(cpu_stat.cpu_times()))
            if self.LOG_MEMORY in self.stats_to_log:
                stats.append('MEM: used/total ')
                stats.append(str(mem_stat.mem_stats()))
            if self.LOG_DISK in self.stats_to_log:
                for disk in self.disks:
                    stats.append('DISK (%s): reads/writes/per second ' % disk)
                    stats.append(str(disk_stat.disk_reads_writes(disk)))
                    stats.append(str(disk_stat.disk_reads_writes_persec(disk)))
            if self.LOG_NETWORK in self.stats_to_log:
                for interface in self.interfaces:
                    stats.append('NET (%s): rx/tx ' % interface)
                    stats.append(str(net_stat.rx_tx_bytes(interface)))
            if stats:
                self.logger.info(' '.join(stats))
        except Exception, e:
            # Don't ever mess up the response because of stats logging
            self.logger.info('The following error occurred while logging '
                             'statistics:  %s' % e)
