#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Models related to OAuth authentication"""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'Token',
    'Consumer',
    'Nonce',
    'DataStore',
    ]

import os.path
import random
import struct
import string
from functools import partial
from logging import getLogger

from oauth.oauth import OAuthToken, OAuthDataStore, OAuthConsumer

from django.contrib.auth.models import User
from django.db import models


TOKEN_LENGTH = 50
TOKEN_SECRET_LENGTH = 50
CONSUMER_SECRET_LENGTH = 30


def _set_seed():
    if (not hasattr(_set_seed, 'seed') and
        os.path.exists("/dev/random")):

        data = open("/dev/random").read(struct.calcsize('Q'))
        random.seed(struct.unpack('Q', data))
        _set_seed.seed = True


def generate_random_string(length):
    _set_seed()
    return ''.join(random.choice(string.ascii_letters)
                   for x in range(length))


class Token(models.Model):
    consumer = models.ForeignKey('Consumer')

    token = models.CharField(
        max_length=TOKEN_LENGTH,
        default=partial(generate_random_string, TOKEN_LENGTH),
        primary_key=True)

    token_secret = models.CharField(
        max_length=TOKEN_SECRET_LENGTH,
        default=partial(generate_random_string, TOKEN_SECRET_LENGTH))

    name = models.CharField(max_length=255, blank=True)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def oauth_token(self):
        """Return OAuthToken with information contained in this model"""
        return OAuthToken(self.token, self.token_secret)

    def __unicode__(self):
        return self.token

    class Meta:
        app_label = 'reviewsapp'


class Consumer(models.Model):
    user = models.OneToOneField(User, related_name='oauth_consumer')

    key = models.CharField(max_length=64)

    secret = models.CharField(max_length=255, blank=True, null=False,
        default=partial(generate_random_string, CONSUMER_SECRET_LENGTH))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    @classmethod
    def lookup_consumer(self, consumer_key):
        # lukasz: To my best knowledge the query below can return more than one
        # entry only if the OpenID provider was switched for already deployed
        # code (between staging/production SSO). That would mean, that the same
        # user on SSO will get second Django user, and second Consumer object,
        # but with exactly the same consumer_key.
        consumers = Consumer.objects.filter(key=consumer_key)
        # If more than one Consumer object is found we choose the one created
        # as the last one keeping up with the assumption of switching the
        # OpenID providers.
        consumers = consumers.order_by('-created_at')
        consumers_count = consumers.count()

        if consumers_count == 0:
            return None
        elif consumers_count > 1:
            getLogger(__name__).info(
                "consumer_key=%s has %d entries in the database",
                consumer_key, consumers_count
            )
        consumer = consumers[0]
        return consumer

    def __unicode__(self):
        return self.key

    def oauth_consumer(self):
        """Return OAuthConsumer based on information contained in this model"""
        return OAuthConsumer(self.key, self.secret)

    class Meta:
        app_label = 'reviewsapp'


class Nonce(models.Model):
    token = models.ForeignKey(Token)
    consumer = models.ForeignKey(Consumer)
    nonce = models.CharField(max_length=255, unique=True, editable=False)
    created_at = models.DateTimeField(auto_now_add=True)

    @classmethod
    def create(cls, consumer_key, token_value, nonce):
        """
        Create new nonce object linked to given consumer and token
        """
        consumer = Consumer.lookup_consumer(consumer_key)
        token = consumer.token_set.get(token=token_value)
        return consumer.nonce_set.create(token=token, nonce=nonce)

    class Meta:
        app_label = 'reviewsapp'

class DataStore(OAuthDataStore):

    def lookup_token(self, token_type, token_field):
        """
        :param token_type: type of token to lookup
        :param token_field: token to look up

        :note: token_type should always be 'access' as only such tokens are
               stored in database

        :returns: OAuthToken object
        """
        assert token_type == 'access'

        try:
            token = Token.objects.get(token=token_field)
            # Piston expects OAuth tokens to have 'consumer' and 'user' atts.
            # (see piston.authentication.OAuthAuthentication.is_authenticated)
            oauthtoken = OAuthToken(token.token, token.token_secret)
            oauthtoken.consumer = token.consumer
            oauthtoken.user = token.consumer.user
            return oauthtoken
        except Token.DoesNotExist:
            return None

    def lookup_consumer(self, consumer_key):
        """
        :param consumer_key: consumer key to lookup

        :returns: OAuthConsumer object
        """
        consumer = Consumer.lookup_consumer(consumer_key)
        if consumer is not None:
            return consumer.oauth_consumer()


    def lookup_nonce(self, consumer, token, nonce):
        """
        :param consumer: OAuthConsumer object
        :param token: OAuthToken object
        :param nonce: nonce to check

        """
        count = Nonce.objects.filter(
            consumer__key=consumer.key,
            token__token=token.key,
            nonce=nonce).count()
        if count > 0:
            return True
        else:
            Nonce.create(consumer.key, token.key, nonce)
            return False

    def __init__(self, oauth_request=None):
        """To serve as a Piston datastore we'll need provide this signature.

        We later won't use the oauth_request, so we can ignore it here.
        """
        return super(DataStore, self).__init__()
