#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Django models."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'Architecture',
    'Repository',
    'Review',
    'ReviewModeration',
    'ReviewModerationFlag',
    'SoftwareItem',
    'SoftwareItemInOrigin',
    'SoftwareItemInRepository',
    ]

from datetime import datetime

from django.contrib.auth.models import (
    Group,
    Permission,
    User,
    )
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Avg, Count
from django.db.models.signals import post_syncdb
from django.utils.translation import ugettext_lazy as _


class ObjectWithReviewStats(models.Model):
    """Mixin for any item which can aggregate review statistics.

    Note: the item must provide a valid reviews_for_object method
    returning a result-set of reviews.
    """
    ratings_total = models.IntegerField(default=0)
    ratings_average = models.DecimalField(default=0, max_digits=3,
                                          decimal_places=2)
    date_ratings_updated = models.DateTimeField(default=datetime.utcnow())

    def update_stats(self):
        stats = self.reviews_for_object().aggregate(
                average=Avg('rating'), total=Count('rating'))
        self.ratings_total = stats['total']

        # If there are no visible reviews, the average will be returned
        # as None.
        if stats['average'] is None:
            stats['average'] = 0

        self.ratings_average = str(stats['average'])
        self.date_ratings_updated = datetime.utcnow()
        self.save()

    class Meta:
        abstract = True


class SoftwareItem(ObjectWithReviewStats):
    """The software item, i.e. package_name/application_name."""
    package_name = models.SlugField(max_length=100)

    def __unicode__(self):
        return u'Package {0}'.format(self.package_name)

    def reviews_for_object(self):
        """See `ObjectWithReviewStats."""
        return Review.objects.filter(softwareitem=self, hide=False)

    class Meta:
        app_label = 'reviewsapp'


class SoftwareItemInRepository(ObjectWithReviewStats):
    """Enables us to record statistics for items in a repo."""

    softwareitem = models.ForeignKey('reviewsapp.SoftwareItem')
    repository = models.ForeignKey('reviewsapp.Repository')

    @property
    def softwareitem_in_origin(self):
        item_in_origin, created = SoftwareItemInOrigin.objects.get_or_create(
            softwareitem=self.softwareitem,
            origin=self.repository.origin)
        return item_in_origin

    def reviews_for_object(self):
        """See `ObjectWithReviewStats."""
        return  Review.objects.filter(
            softwareitem=self.softwareitem, repository=self.repository,
            hide=False)

    def update_stats(self):
        """Overridden to ensure that all stats are updated at once."""
        super(SoftwareItemInRepository, self).update_stats()
        self.softwareitem.update_stats()
        self.softwareitem_in_origin.update_stats()

    class Meta:
        app_label = 'reviewsapp'
        unique_together = (('softwareitem', 'repository'),)


class SoftwareItemInOrigin(ObjectWithReviewStats):
    """Enables us to record statistics for items in a repo."""

    softwareitem = models.ForeignKey('reviewsapp.SoftwareItem')
    origin = models.SlugField(max_length=100, db_index=True)

    def reviews_for_object(self):
        """See `ObjectWithReviewStats."""
        return Review.objects.filter(
            softwareitem=self.softwareitem, repository__origin=self.origin,
            hide=False)

    class Meta:
        app_label = 'reviewsapp'
        unique_together = (('softwareitem', 'origin'),)


class Repository(models.Model):
    """The repository a package comes from.

    That includes the origin (e.g. ubuntu) and the distro release
    (e.g. lucid).
    """
    # Ubuntu, ppa-ppa-name, ...
    origin = models.SlugField(max_length=100)
    # karmic, lucid
    distroseries = models.SlugField(max_length=25)

    def __unicode__(self):
        return u'{0}/{1}'.format(self.distroseries, self.origin)

    class Meta:
        verbose_name_plural = 'Repositories'
        app_label = 'reviewsapp'


class ReviewModerationManager(models.Manager):
    """Provide custom filters based on the status."""
    use_for_related_fields = True

    def pending(self):
        return self.get_query_set().filter(
            status=ReviewModeration.PENDING_STATUS)

    def approved(self):
        return self.get_query_set().filter(
            status=ReviewModeration.APPROVED_STATUS)

    class Meta:
        app_label = 'reviewsapp'


class ReviewModeration(models.Model):
    """A moderated object.

    Record the pending moderation and the action taken.

    During active moderation (moderate by default) a pending
    ReviewModeration may be created when items are first saved.

    Afterwards, or during passive moderation (moderate when flagged),
    a ModerationObject may be created when an item is flagged.
    """
    PENDING_STATUS = 0
    APPROVED_STATUS = 1
    REJECTED_STATUS = 2
    MODERATION_CHOICES = (
        (APPROVED_STATUS, _('Approved')),
        (REJECTED_STATUS, _('Rejected')),
        )
    STATUS_CHOICES = MODERATION_CHOICES + (
        (PENDING_STATUS, _('Pending')),
        )

    review = models.ForeignKey('Review')

    status = models.SmallIntegerField(
        choices=STATUS_CHOICES, default=PENDING_STATUS)

    date_created = models.DateTimeField(default=datetime.utcnow,
        editable=False)
    date_moderated = models.DateTimeField(null=True, blank=True)
    moderator = models.ForeignKey(User, null=True, blank=True)
    moderation_text = models.TextField(blank=True)

    objects = ReviewModerationManager()

    def moderate(self, moderator, status, moderation_text):
        """Record the moderation of the review and hide/un-hide it."""
        if not self.can_be_moderated_by(moderator):
            raise ReviewModeration.ModerationError(
                "This item cannot be moderated by '{0}'.".format(
                    moderator.username))

        if self.status != ReviewModeration.PENDING_STATUS:
            raise ReviewModeration.ModerationError(
                "Only pending moderations can be moderated.")
        if status == ReviewModeration.PENDING_STATUS:
            raise ReviewModeration.ModerationError(
                "You must approve or reject to moderate.")

        self.moderator = moderator
        self.status = status
        self.moderation_text = moderation_text
        self.date_moderated = datetime.utcnow()
        self.save()

        # Hide or unhide the related review as appropriate.
        should_hide = (status == ReviewModeration.REJECTED_STATUS)
        if self.review.hide != should_hide:
            self.review.hide = should_hide
            self.review.save()
            self.review.softwareitem_in_repository.update_stats()

    def can_be_moderated_by(self, user):
        """Return whether the user is allowed to moderate this object."""
        if self.review.reviewer == user:
            return False

        if self.reviewmoderationflag_set.filter(user=user):
            return False

        return True

    class Meta:
        app_label = 'reviewsapp'

    class ModerationError(Exception):
        """Raised when ReviewModeration.moderate is called incorrectly."""


class ReviewModerationFlag(models.Model):
    """An object that has been flagged for moderation."""
    review_moderation = models.ForeignKey(ReviewModeration)
    summary = models.CharField(max_length=200)
    description = models.TextField()
    date_created = models.DateTimeField(default=datetime.utcnow,
        editable=False)
    user = models.ForeignKey(User, null=True)

    class Meta:
        ordering = ['-date_created']
        app_label = 'reviewsapp'

    class UserAlreadyFlaggedReview(Exception):
        """Raised when a user has already flagged a certain review once."""


class Architecture(models.Model):
    """Represents a machine architecture.

    Note: This is only included as a model (rather than a char attribute
    on Review) so that we can dynamically update the list of valid
    architectures (either by adding them via the admiin interface or by
    checking with Launchpad).
    """
    tag = models.CharField(unique=True, max_length=16)

    class Meta:
        app_label = 'reviewsapp'

    def __unicode__(self):
        return self.tag


class Review(models.Model):
    """A specific review."""
    softwareitem = models.ForeignKey(SoftwareItem)
    version = models.CharField(max_length=100)
    architecture = models.ForeignKey(Architecture)
    repository = models.ForeignKey(Repository)
    reviewer = models.ForeignKey(User)
    date_created = models.DateTimeField(default=datetime.utcnow,
        editable=False)
    rating = models.IntegerField()
    summary = models.CharField(max_length=80)
    review_text = models.CharField(max_length=5000)
    language = models.SlugField(max_length=10)
    hide = models.BooleanField(
        'Hide review from the UI', db_index=True, default=False)
    # The following get updated when a usefulness vote is added
    usefulness_total = models.IntegerField(default=0)
    usefulness_favorable = models.IntegerField(default=0)
    usefulness_percentage = models.IntegerField(default=0)
    # XXX 2011-03-18 michaeln Remove application_name eventually.
    # After removing the app_name attribute of software item, we add it
    # here on the review so that we're not losing any information being
    # passed by the client.
    app_name = models.CharField(max_length=100, default='', blank=True)

    def __unicode__(self):
        return u'lang: {0}, rating: {1}, summary: {2}\n{3}'.format(
            self.language, self.rating, self.summary, self.review_text)

    @property
    def softwareitem_in_repository(self):
        """Helper property until we switch the foreignkey."""
        item_repo, created = SoftwareItemInRepository.objects.get_or_create(
            softwareitem=self.softwareitem, repository=self.repository)
        return item_repo

    def package_name(self):
        return self.softwareitem.package_name

    def origin(self):
        return self.repository.origin

    def distroseries(self):
        return self.repository.distroseries

    def package_app_name(self):
        app_name = self.app_name
        if app_name:
            return self.package_name() + "/" + app_name
        else:
            return self.package_name()

    def reviewer_username(self):
        return self.reviewer.username

    def reviewer_displayname(self):
        return self.reviewer.get_full_name()

    def flag(self, user, summary, description):
        """Flag an object as possibly requiring moderation.

        @param user: The user flagging the object.
        @param summary: A summary identifying the reason for flagging the
            item such as 'Spam', or 'Offensive language'.
        @param description: A description of the reason for flagging the
            item.

        # XXX michaeln 2010-11-29 Extract moderation functionality.
        This method may be added automatically when registering an
        item.
        """
        # Ensure that this user hasn't already flagged this item.
        previous_flags_for_user = ReviewModerationFlag.objects.filter(
            review_moderation__review=self,
            user=user)
        if previous_flags_for_user.count() > 0:
            raise ReviewModerationFlag.UserAlreadyFlaggedReview()

        review_moderation, created = self.reviewmoderation_set.get_or_create(
            status=ReviewModeration.PENDING_STATUS)
        flagged_object = review_moderation.reviewmoderationflag_set.create(
            user=user, summary=summary, description=description)

        # If this object hasn't been approved previously, then we hide
        # it.
        if self.reviewmoderation_set.approved().count() == 0:
            self.hide = True
            self.save()
            self.softwareitem_in_repository.update_stats()

        return flagged_object

    class Meta:
        app_label = 'reviewsapp'


class Usefulness(models.Model):
    review = models.ForeignKey(Review, db_index=True)
    user = models.ForeignKey(User)
    useful = models.BooleanField(db_index=True)

    def username(self):
        return self.user.username

    def review_id(self):
        return self.review.id

    class Meta:
        app_label = 'reviewsapp'



class RNRSettings(models.Model):
    PASSIVE_MODERATION = 'passive'
    ACTIVE_MODERATION = 'active'
    BLACKLIST_MODERATION = 'blcklst'
    MODERATION_CHOICES = (
        (PASSIVE_MODERATION, 'Passive - automated approval'),
        (ACTIVE_MODERATION, 'Active - manual approval'),
        (BLACKLIST_MODERATION, 'Automated moderation based on blacklist'),
    )

    moderation_mode = models.CharField(default=PASSIVE_MODERATION,
        max_length=8, choices=MODERATION_CHOICES)

    blacklist_words = models.TextField(default='',
        help_text=_("Enter one word per line."))

    @classmethod
    def get_settings(cls):
        # There is only ever a singleton setting object.
        try:
            settings = RNRSettings.objects.get()
        except RNRSettings.DoesNotExist:
            settings = RNRSettings.objects.create()

        return settings

    @classmethod
    def set_moderation_mode(cls, mode):
        settings = cls.get_settings()
        settings.moderation_mode = mode
        settings.save()

    @classmethod
    def should_moderate_text(cls, *text):
        settings = cls.get_settings()
        if settings.moderation_mode == cls.ACTIVE_MODERATION:
            return True
        elif settings.moderation_mode == cls.PASSIVE_MODERATION:
            return False

        blacklist_words = settings.blacklist_words.split('\n')
        combined_texts = "\n".join(text)
        for blacklist_word in blacklist_words:
            if blacklist_word in combined_texts:
                return True

        return False

    @classmethod
    def get_moderation_summary_and_desc(cls):
        moderation_mode = cls.get_settings().moderation_mode
        if moderation_mode == cls.ACTIVE_MODERATION:
            return (
                'Automatically Flagged',
                'This review was automatically flagged because the rnr '
                'server was operating in active moderation mode')
        elif moderation_mode == cls.BLACKLIST_MODERATION:
            return (
                'Contains Blacklisted Word',
                'This review was flagged because the server was operating '
                'in blacklist mode and the review contained a blacklisted '
                'word.')
        else:
            return None, None

    class Meta:
        app_label = 'reviewsapp'
        verbose_name_plural = 'RNR Settings'


# GroupPermissions can't go in fixtures because Permissions are stored as
# model metadata, so our Permission's primary key can change without warning
def post_syncdb_handler(sender, **kwargs):
    if sender.__name__ == 'reviewsapp.models':
        ct = ContentType.objects.get_for_model(ReviewModeration)
        perm, created = Permission.objects.get_or_create(
            codename='can_moderate',
            name="Can resolve moderation requests",
            content_type=ct,
        )
        group, created = Group.objects.get_or_create(name='rnr-moderators')
        if not perm in group.permissions.all():
            group.permissions.add(perm)
            group.save()
        group, created = Group.objects.get_or_create(name='rnr-developers')

post_syncdb.connect(post_syncdb_handler)
