#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A TestCase class with a built in object factory."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'ReviewsObjectFactory',
    'TestCaseWithFactory',
    ]


import os
from datetime import datetime
from itertools import count

from django.contrib.auth.models import User
from django.test import (
    TestCase,
    TransactionTestCase,
    )
from django_openid_auth.models import UserOpenID
from testtools import TestCase as TestToolsTestCase

from reviewsapp.models import (
    Architecture,
    Repository,
    Review,
    ReviewModeration,
    SoftwareItem,
    SoftwareItemInOrigin,
    SoftwareItemInRepository,
    Token,
    Consumer,
    Usefulness
    )
from reviewsapp.utilities import full_claimed_id


class RequestTestObject(object):
    """Test object for emulating GET requests"""
    def __init__(self,fakeGetDict=None):
        if fakeGetDict:
            self.GET = fakeGetDict
        else:
            self.GET = {}


class ReviewsObjectFactory(object):
    """A factory for creating model objects for testing."""

    app_reviews_counter = count()

    def getUniqueInteger(self):
        """Return a unique integer for the test run."""
        return self.app_reviews_counter.next()

    def getUniqueString(self, prefix=None):
        if prefix is None:
            prefix = "generic-string"
        return prefix + str(self.getUniqueInteger())

    def makeUser(self, username=None, email=None, password='test',
        first_name=None, last_name=None, open_id=None, is_admin=False):
        if username is None:
            username = self.getUniqueString(prefix='username')
        if email is None:
            email = self.getUniqueString(prefix='email') + '@example.com'
        if first_name is None:
            first_name = self.getUniqueString(prefix='firstname')
        if last_name is None:
            last_name = self.getUniqueString(prefix='lastname')

        user = User.objects.create_user(
            username=username, email=email, password=password)
        user.first_name = first_name
        user.last_name = last_name

        if is_admin:
            user.is_staff = True
            user.is_superuser = True
        user.save()

        # Create an openid record too.
        if open_id is None:
            open_id = full_claimed_id(self.getUniqueString(prefix='ident-'))
        elif open_id is False:
            return user
        useropenid = UserOpenID.objects.create(
            user=user, claimed_id=open_id, display_id=open_id)

        return user

    def makeVersion(self):
        return '{0}.0'.format(self.getUniqueInteger())

    def makeSoftwareItem(self, package_name=None,
                         ratings_total=1, ratings_average=3.5,
                         date_ratings_updated=None):
        if package_name is None:
            package_name = self.getUniqueString(prefix='package')
        if date_ratings_updated == None:
            date_ratings_updated = datetime.utcnow()

        return SoftwareItem.objects.create(package_name=package_name,
            ratings_total=ratings_total,
            ratings_average=unicode(ratings_average),
            date_ratings_updated=date_ratings_updated)

    def makeSoftwareItemInRepository(self, software_item=None,
                                     repository=None):
        if software_item is None:
            software_item = self.makeSoftwareItem()
        if repository is None:
            repository = self.makeRepository()

        return SoftwareItemInRepository.objects.create(
                softwareitem=software_item, repository=repository)

    def makeSoftwareItemInOrigin(self, software_item=None,
                                 origin=None):
        if software_item is None:
            software_item = self.makeSoftwareItem()
        if origin is None:
            origin = self.getUniqueString(prefix='origin')

        return SoftwareItemInOrigin.objects.create(
                softwareitem=software_item, origin=origin)

    def makeRepository(self, origin='ubuntu', distro_series=None):
        if distro_series is None:
            distro_series = self.getUniqueString(prefix='series')

        return Repository.objects.create(
            origin=origin, distroseries=distro_series)

    def makeUsefulness(self, review=None, user=None, useful=True):
        if review is None:
            review = self.makeReview()
        if user is None:
            user = self.makeUser()

        usefulness = Usefulness.objects.create(review=review, user=user,
                        useful=useful)

        return usefulness

    def makeReview(self, reviewer=None, software_item=None, repository=None,
                   version=None, rating="5", summary=None, review_text=None,
                   date_created=None, language=None, hide=False, num_flags=0,
                   architecture='i386', app_name=''):
        if reviewer is None:
            reviewer = self.makeUser()
        if software_item is None:
            software_item = self.makeSoftwareItem()
        if repository is None:
            repository = self.makeRepository()
        if version is None:
            version = self.makeVersion()
        if summary is None:
            summary = self.getUniqueString(prefix='summary')
        if review_text is None:
            review_text = self.getUniqueString(prefix='review_text')
        if date_created is None:
            date_created = datetime.utcnow()
        if language is None:
            language = self.getUniqueString(prefix='lang')

        architecture, created = Architecture.objects.get_or_create(
            tag=architecture)

        review = Review.objects.create(reviewer=reviewer,
            softwareitem=software_item, repository=repository,
            version=version, rating=rating, summary=summary,
            review_text=review_text, language=language, hide=hide,
            date_created=date_created, architecture=architecture,
            app_name=app_name)
        review.softwareitem_in_repository.update_stats()
        for flag_count in range(num_flags):
            review.flag(
                self.makeUser(),
                self.getUniqueString(prefix='flag-summary'),
                self.getUniqueString(prefix='flag-description'))
        return review

    def makeReviewModeration(self, num_flags=1,
                             status=ReviewModeration.PENDING_STATUS,
                             language=None, distro_series=None):
        repository = self.makeRepository(distro_series=distro_series)
        review = self.makeReview(language=language, repository=repository)
        review_moderation = ReviewModeration.objects.create(review=review)
        for flag_count in range(num_flags):
            review.flag(
                self.makeUser(),
                self.getUniqueString(prefix='flag-summary'),
                self.getUniqueString(prefix='flag-description'))

        if status != ReviewModeration.PENDING_STATUS:
            review_moderation.status = status
            review_moderation.save()

        return review_moderation

    def makeOAuthTokenAndConsumer(self, user=None, consumer_secret=None,
            token_string=None, token_secret=None, save=True):
        """Create a new set of OAuth token and consumer creds."""
        if user is None:
            user = self.makeUser()
        if token_string is None:
            token_string = self.getUniqueString(prefix='token')
        if token_secret is None:
            token_secret = self.getUniqueString(prefix='token-secret')
        if consumer_secret is None:
            consumer_secret = self.getUniqueString(prefix='consumer-secret')

        try:
            openid = user.useropenid_set.get()
        except UserOpenID.DoesNotExist:
            open_id = full_claimed_id(self.getUniqueString(prefix='ident-'))
            openid = UserOpenID.objects.create(
                user=user, claimed_id=open_id, display_id=open_id)
        consumer_key = openid.claimed_id.split('/')[-1]
        consumer = Consumer(user=user, key=consumer_key,
            secret=consumer_secret)
        if save:
            consumer.save()
        token_name = self.getUniqueString(prefix='token-name')
        token = Token(consumer=consumer, token=token_string,
            token_secret=token_secret, name=token_name)
        if save:
            token.save()
        return token, consumer

    def get_test_file(self, file_name):
        return os.path.join(os.path.dirname(__file__), 'test_data',
                            file_name)


class TestCaseWithFactory(TestCase, TestToolsTestCase):

    def setUp(self):
        super(TestCaseWithFactory, self).setUp()
        self.factory = ReviewsObjectFactory()


class TxTestCaseWithFactory(TransactionTestCase, TestToolsTestCase):

    def setUp(self):
        super(TxTestCaseWithFactory, self).setUp()
        self.factory = ReviewsObjectFactory()

    def tearDown(self):
        super(TxTestCaseWithFactory, self).tearDown()
