#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Ubuntu App Review auth tests."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'SSOOAuthAuthenticationTestCase',
]

from django.conf import settings
from django.contrib.auth.models import User
from django.http import HttpResponse
import httplib
from mock import patch
from oauth.oauth import (OAuthRequest, OAuthConsumer, OAuthToken,
    OAuthSignatureMethod_PLAINTEXT)
from piston.resource import Resource
from piston.handler import BaseHandler

from reviewsapp.auth import SSOOAuthAuthentication
from reviewsapp.models import Consumer
from reviewsapp.tests.factory import TestCaseWithFactory
from reviewsapp.tests.helpers import mock_data_for_account


class MockRequest:
    def __init__(self, method='GET'):
        self.META = {'QUERY_STRING': ''}
        self.GET = {}
        self.POST = {}
        self.REQUEST = {}
        self.method = method
        self.environ = self.META

    def build_absolute_uri(self):
        return 'http://localhost:8000/api/1.0/foo'


class MockGetHandler(BaseHandler):
    allowed_methods = ('GET',)
    def read(self, request):
        self.called = True
        return 'Hi!'


class MockPostHandler(BaseHandler):
    allowed_methods = ('POST',)
    def create(self, request):
        self.called = True
        return HttpResponse(status=httplib.CREATED)


class SSOOAuthAuthenticationTestCase(TestCaseWithFactory):
    def setUp(self):
        self.auth = SSOOAuthAuthentication(realm="Ratings and Reviews")
        self.get_resource = Resource(MockGetHandler,
            authentication=self.auth)
        self.post_resource = Resource(MockPostHandler,
            authentication=self.auth)
        super(SSOOAuthAuthenticationTestCase, self).setUp()

    def header_from_token(self, request, token, consumer):
        consumer = OAuthConsumer(consumer.key, consumer.secret)
        token = OAuthToken(token.token, token.token_secret)
        oauth_request = OAuthRequest.from_consumer_and_token(
            consumer, token, http_url=request.build_absolute_uri())
        oauth_request.sign_request(OAuthSignatureMethod_PLAINTEXT(),
            consumer, token)
        headers = oauth_request.to_header('Ratings and Reviews')
        return {'HTTP_AUTHORIZATION': headers['Authorization']}

    def test_no_auth_fails(self):

        response = self.get_resource(MockRequest())

        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_success(self, mock_get_data):
        request, token, consumer = self.make_request_with_token()

        response = self.get_resource(request)

        self.assertFalse(mock_get_data.called)
        self.assertEqual(httplib.OK, response.status_code)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_fetch_creds_from_sso(self, mock_get_data):
        request, token, consumer = self.make_request_with_token(save=False)
        mock_get_data.return_value = mock_data_for_account(consumer, token)

        response = self.get_resource(request)

        self.assertTrue(mock_get_data.called)
        self.assertEqual(httplib.OK, response.status_code)

    @patch('reviewsapp.auth.initialize_server_request')
    def test_initialize_server_request_fails(self, mock_init_request):
        """Server shouldn't return a 500 if initialize_server_request fails"""
        mock_init_request.return_value = None, None
        request, token, consumer = self.make_request_with_token()

        response = self.get_resource(request)

        self.assertEqual(httplib.UNAUTHORIZED, response.status_code)

    def test_post_data_doesnt_replace_parameters(self):
        """Check that POST data doesn't replace information from
        the Authentication header in the OAuthRequest parameter set.
        If the post request succeeds then the information from the header
        made it into the parameter list.
        """
        request, token, consumer = self.make_request_with_token(method='POST')
        request.POST = request.REQUEST = {'foo': 'bar', 'baz': 'troz'}

        response = self.post_resource(request)

        self.assertEqual(httplib.CREATED, response.status_code)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_unknown_token(self, mock_get_data):
        """Check that if the consumer is known, but a new token is provided
        the data is fetched from SSO
        """
        request, token, consumer = self.make_request_with_token(method='POST')
        account_data = mock_data_for_account(consumer, token)
        orig_plaintext_sig = "%s&%s" % (
            account_data['consumer_secret'], account_data['token_secret'])
        mock_get_data.return_value = account_data
        request.POST = request.REQUEST = {'foo': 'bar', 'baz': 'troz'}
        orig_token = token.token
        token.delete()

        response = self.post_resource(request)

        mock_get_data.assert_called_with(token=orig_token,
            openid_identifier=consumer.key, signature=orig_plaintext_sig)
        self.assertEqual(httplib.CREATED, response.status_code)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_user_doesnt_exist(self, mock_get_data):
        """Check that a user is created if none exists"""
        request, token, consumer = self.make_request_with_token(save=False)
        data_for_account = mock_data_for_account(consumer, token)
        mock_get_data.return_value = data_for_account
        consumer.user.delete()
        request = MockRequest()
        request.META.update(self.header_from_token(request, token, consumer))
        new_user = self.factory.makeUser()

        with patch('django.contrib.auth.models.UserManager.create_user') as \
                   mock_create_user:
            mock_create_user.return_value = new_user
            response = self.get_resource(request)

        mock_create_user.assert_called_with(data_for_account['username'],
            data_for_account['preferred_email'], password=None)
        self.assertEqual(httplib.OK, response.status_code)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_user_doesnt_exist_twice(self, mock_get_data):
        """Check that an autocreated user persists on for a second call"""
        request, token, consumer = self.make_request_with_token(save=False)
        mock_get_data.return_value = mock_data_for_account(consumer, token)
        consumer.user.delete()
        new_user = self.factory.makeUser()
        create_user = 'django.contrib.auth.models.UserManager.create_user'
        with patch(create_user) as mock_create_user:
            mock_create_user.return_value = new_user
            # First request...
            response = self.get_resource(request)
            request = MockRequest()
            request.META.update(self.header_from_token(request, token,
                consumer))

            response = self.get_resource(request)

        self.assertEqual(1, mock_create_user.call_count)
        self.assertEqual(1, mock_get_data.call_count)
        self.assertEqual(httplib.OK, response.status_code)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_user_is_created_with_right_claimed_id(self, mock_get_data):
        """Check that autocreated users get the right (full) claimed_id"""
        request, token, consumer = self.make_request_with_token(save=False)
        data_for_account = mock_data_for_account(consumer, token)
        mock_get_data.return_value = data_for_account
        consumer.user.delete()
        new_user = self.factory.makeUser()
        new_user.useropenid_set.all().delete()

        create_user = 'django.contrib.auth.models.UserManager.create_user'
        with patch(create_user) as mock_create_user:
            mock_create_user.return_value = new_user
            response = self.get_resource(request)

        self.assertEqual(1, new_user.useropenid_set.count())
        claimed_id = new_user.useropenid_set.get().claimed_id
        self.assertEqual(claimed_id, settings.OPENID_SSO_SERVER_URL + '+id/' +
            consumer.key)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_stale_consumer(self, mock_get_data):
        """Check that having a stale stored Consumer doesn't break auth"""
        user = self.factory.makeUser()
        # create a (stale) consumer
        self.factory.makeOAuthTokenAndConsumer(user=user)
        # and a new one
        token, consumer = self.factory.makeOAuthTokenAndConsumer(user=user,
            save=False)
        request = MockRequest()
        request.META.update(self.header_from_token(request, token, consumer))
        mock_get_data.return_value = mock_data_for_account(consumer, token)

        response = self.get_resource(request)

        # The original consumer has been replaced with the new one from SSO.
        self.assertEqual(1, Consumer.objects.filter(key=consumer.key).count())
        self.assertEqual(consumer.secret,
            Consumer.objects.get(key=consumer.key).secret)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_first_and_last_name_updated(self, mock_get_data):
        """Check that the users name is updated from SSO."""
        user = self.factory.makeUser(first_name='Harry', last_name='Jones')
        # create a (stale) consumer
        self.factory.makeOAuthTokenAndConsumer(user=user)
        # and a new one
        token, consumer = self.factory.makeOAuthTokenAndConsumer(user=user,
            save=False)
        request = MockRequest()
        request.META.update(self.header_from_token(request, token, consumer))
        mock_get_data.return_value = mock_data_for_account(consumer, token,
            displayname='Hazel Jennet')

        response = self.get_resource(request)

        # The original first and last names have been updated from SSO.
        user = User.objects.get(id=user.id)
        self.assertEqual('Hazel Jennet', user.get_full_name())

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_auth_succeedes_when_more_than_one_consumer(self, mock_get_data):
        """
        Check that auth works even when there are two consumers for one token

        Bug: #706035
        
        """
        user1 = self.factory.makeUser()
        user2 = self.factory.makeUser()

        token1, consumer1 = self.factory.makeOAuthTokenAndConsumer(user=user1)
        token2, consumer2 = self.factory.makeOAuthTokenAndConsumer(user=user2)
        consumer2.key = consumer1.key
        consumer2.save()

        request = MockRequest()
        request.META.update(
            self.header_from_token(request, token2, consumer2))
        mock_get_data.return_value = mock_data_for_account(consumer2, token2)

        response = self.get_resource(request)
        self.assertEquals(response.status_code, 200)

    def make_request_with_token(self, save=True, method='GET'):
        token, consumer = self.factory.makeOAuthTokenAndConsumer(save=save)
        request = MockRequest(method=method)
        request.META.update(self.header_from_token(request, token, consumer))
        return request, token, consumer
