#   -*- coding: utf-8 -*-
#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Ubuntu App Review handlers tests."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'FlagReviewHandlerTestCase',
    'ReviewsStatsHandlerTestCase',
    'ServerStatusHandlerTestCase',
    'ShowReviewsHandlerTestCase',
    'SubmitReviewHandlerTestCase',
    'UsefulnessHandlerTestCase',
    'ShowUsefulnessHandlerTestCase',
    ]


import httplib
from datetime import datetime, timedelta
from urllib import quote_plus

from mock import Mock, patch
from django.conf import settings
from django.core.cache import cache
from django.core.urlresolvers import (
    NoReverseMatch,
    reverse,
    )
from django.test import TestCase
from django.utils import simplejson

from reviewsapp.api.handlers import (
    FlagReviewHandler,
    ServerStatusHandler,
    ShowReviewsHandler,
    SubmitUsefulnessHandler,
    ShowUsefulnessHandler,
)
from reviewsapp.models import (
    Repository,
    Review,
    ReviewModerationFlag,
    SoftwareItem,
    )
from reviewsapp.tests.factory import TestCaseWithFactory
from reviewsapp.tests.factory import RequestTestObject
from reviewsapp.tests.helpers import patch_settings
from reviewsapp.tests.matchers import (
    ContainsStatsForPackage,
)
from reviewsapp.utilities import cache_key_for_url

def _get_url_kwargs_dict_for_review(review):
    # this is the quoting we expect
    appname=quote_plus(quote_plus(
        ";"+review.app_name.encode("utf-8")))
    pkgname=review.softwareitem.package_name
    return dict(language=review.language,
                origin=review.repository.origin,
                distroseries=review.repository.distroseries,
                package_name="%s%s" % (pkgname, appname),
                version=review.version)

class ServerStatusHandlerTestCase(TestCase):
    def test_read(self):
        ss_handler = ServerStatusHandler()
        self.assertEqual('ok', ss_handler.read(None))

    def test_request_not_cached(self):
        # Requests for server status are never cached.
        response = self.client.get(reverse('rnr-server-status'))

        self.assertEqual('max-age=0', response['cache-control'])

        # A subsequent call will hit the view.
        with patch('reviewsapp.api.handlers.ServerStatusHandler.read') as (
            mock_read):
            response = self.client.get(reverse('rnr-server-status'))
            self.assertTrue(mock_read.called)


class ReviewsStatsHandlerTestCase(TestCaseWithFactory):

    def _request_stats(self, last_days=0, data=None, origin='any',
                       distroseries='any', use_old_api=False):
        data = data or {}
        if use_old_api:
            kwargs = {}
        else:
            kwargs = dict(origin=origin, distroseries=distroseries)
        if last_days > 0:
            kwargs['days'] = last_days
        url = reverse('rnr-api-reviews-stats', kwargs=kwargs)
        return self.client.get(url, data)

    def test_read(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='de')
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'),
            repository=self.factory.makeRepository('ubuntu', 'natty'),
            language='de')

        response = self._request_stats()
        reviews = simplejson.loads(response.content)

        self.assertEqual(2, len(reviews))
        self.assertThat(reviews, ContainsStatsForPackage(
                'package_foo', average=5, count=1))
        self.assertThat(reviews, ContainsStatsForPackage(
                'package_bar', average=5, count=1))

    def test_read_for_repo(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='de')
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'),
            repository=self.factory.makeRepository('ubuntu', 'natty'),
            language='de')

        response = self._request_stats(origin='ubuntu', distroseries='hardy')
        reviews = simplejson.loads(response.content)

        self.assertEqual(1, len(reviews))
        self.assertThat(reviews, ContainsStatsForPackage(
                'package_foo', average=5, count=1))

    def test_read_for_origin(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='de')
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'),
            repository=self.factory.makeRepository('canonical', 'natty'),
            language='de')

        response = self._request_stats(origin='ubuntu', distroseries='any')
        reviews = simplejson.loads(response.content)

        self.assertEqual(1, len(reviews))
        self.assertThat(reviews, ContainsStatsForPackage(
                'package_foo', average=5, count=1))


    def test_request_cached_with_extra_headers(self):
        # Requests for server status are never cached.
        response = self._request_stats()

        self.assertEqual(
            'max-age={0}'.format(settings.CACHE_REVIEW_STATS_SECONDS),
            response['cache-control'])

        # A subsequent call will not hit the view.
        read_handler = 'reviewsapp.api.handlers.ReviewsStatsHandler.read'
        with patch(read_handler) as mock_read:
            response = self._request_stats()
            self.assertFalse(mock_read.called)

    def test_cannot_request_with_get_params(self):
        # Requesting the un-cached response (by adding get params) is
        # not allowed.
        response = self._request_stats(data=dict(foo='bar'))

        self.assertEqual(400, response.status_code)

    def test_updates_last_days(self):
        # Only items which have updated stats are returned.
        stat_4_days_old = self.factory.makeSoftwareItem(
            date_ratings_updated=datetime.utcnow() - timedelta(4))
        stat_2_days_old = self.factory.makeSoftwareItem(
            date_ratings_updated=datetime.utcnow() - timedelta(2))
        stat_0_days_old = self.factory.makeSoftwareItem()

        response = self._request_stats(last_days=3)

        response_decoded = simplejson.loads(response.content)
        self.assertEqual(2, len(response_decoded))
        self.assertThat(response_decoded, ContainsStatsForPackage(
            stat_2_days_old.package_name, average='3.5', count=1))
        self.assertThat(response_decoded, ContainsStatsForPackage(
            stat_0_days_old.package_name, average='3.5', count=1))

    def test_updates_last_days_matches_only_specified_days(self):
        # The handler cannot be used to generate random updates.
        # The following are all valid.
        kwargs = dict(origin='ubuntu', distroseries='natty')
        for days in (1, 3, 7):
            kwargs['days'] = days
            reverse('rnr-api-reviews-stats', kwargs=kwargs)

        # But other options for days will not hit the handler.
        for days in (2, 4, 8):
            kwargs['days'] = days
            self.assertRaises(
                NoReverseMatch, reverse, 'rnr-api-reviews-stats',
                kwargs=kwargs)

    def test_stats_old_api(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'))

        response = self._request_stats(use_old_api=True)
        reviews = simplejson.loads(response.content)

        self.assertEqual(1, len(reviews))
        self.assertThat(reviews, ContainsStatsForPackage(
                'package_foo', average=5, count=1))

    def test_stats_old_api_with_days(self):
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'hardy'))

        response = self._request_stats(last_days=1, use_old_api=True)
        reviews = simplejson.loads(response.content)

        self.assertEqual(1, len(reviews))
        self.assertThat(reviews, ContainsStatsForPackage(
                'package_foo', average=5, count=1))

    def test_read_across_origins(self):
        # It is possible to request natty apps from various origins
        # (ubuntu, a ppa, or canonical etc.)
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            repository=self.factory.makeRepository('ubuntu', 'natty'),
            language='de')
        self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'),
            repository=self.factory.makeRepository('canonical', 'natty'),
            language='de')

        response = self._request_stats(origin='any', distroseries='natty')
        reviews = simplejson.loads(response.content)

        self.assertEqual(2, len(reviews))
        self.assertThat(reviews, ContainsStatsForPackage(
                'package_foo', average=5, count=1))
        self.assertThat(reviews, ContainsStatsForPackage(
                'package_bar', average=5, count=1))


class ShowUsefulnessHandlerTestCase(TestCaseWithFactory):
    def test_read_for_single_parameter(self):
        '''test that a single usefulness vote by a user is returned when
         searching for that user'''
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'))
        user = self.factory.makeUser()
        usefulness = self.factory.makeUsefulness(review, user)

        handler = ShowUsefulnessHandler()

        request1 = RequestTestObject({'username': user.username})
        request2 = RequestTestObject({'review_id' : review.id})

        results1 = handler.read(request1)
        results2 = handler.read(request2)

        self.assertEqual(1, len(results1))
        self.assertEqual(usefulness, results1[0])
        self.assertEqual(usefulness, results2[0])

    def test_read_filtering(self):
        '''test that a passed parameters results in all relevant items
        being returned, but no unwanted items'''
        review1 = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'))
        review2 = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('foo_bar'))
        user1 = self.factory.makeUser()
        user2 = self.factory.makeUser()
        usefulness1 = self.factory.makeUsefulness(review1, user1)
        usefulness2 = self.factory.makeUsefulness(review1, user2)
        usefulness3 = self.factory.makeUsefulness(review2, user1)

        handler = ShowUsefulnessHandler()

        request1 = RequestTestObject({'username':user1.username})
        request2 = RequestTestObject({'review_id':review1.id})
        request3 = RequestTestObject(
            {'username':user1.username, 'review_id':review2.id}
            )

        results1 = handler.read(request1)
        results2 = handler.read(request2)
        results3 = handler.read(request3)

        #test username filter results
        self.assertEqual(2, len(results1))
        self.assertIn(usefulness1, results1)
        self.assertIn(usefulness3, results1)
        self.assertNotIn(usefulness2, results1)

        #test review_id filter results
        self.assertEqual(2, len(results2))
        self.assertIn(usefulness1, results2)
        self.assertIn(usefulness2, results2)
        self.assertNotIn(usefulness3, results2)

        #test usefulness AND review_id filter results
        self.assertEqual(1, len(results3))
        self.assertNotIn(usefulness1, results3)
        self.assertNotIn(usefulness2, results3)
        self.assertIn(usefulness3, results3)


    def test_read_for_no_parameters_gives_400_error(self):
        '''test that a http 400 error is returned if no valid parameter
        is passed'''
        handler = ShowUsefulnessHandler()
        request1 = RequestTestObject()
        request2 = RequestTestObject({'something':'nothing'})
        results1 = handler.read(request1)
        results2 = handler.read(request2)
        self.assertEqual(httplib.BAD_REQUEST, results1.status_code)
        self.assertEqual(httplib.BAD_REQUEST, results2.status_code)


class ShowReviewsHandlerTestCase(TestCaseWithFactory):

    def test_review_for_package(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'))
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'any', 'any', 'any', 'any',
            'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertEqual(review, reviews[0])

    def test_review_for_app(self):
        # We no-longer differentiate for app_names, returning all
        # reviews for the package instead.
        software_item = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=software_item,
            app_name='app_foo')
        review2 = self.factory.makeReview(software_item=software_item,
            app_name='AnotherApp')
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'any', 'any', 'any', 'any',
                               'package_foo;app_foo')

        self.assertEqual(set([review1, review2]), set(reviews))

    def test_review_for_app_url(self):
        # test utf8 encoded app
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            app_name=u'Déjà Dup')
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=_get_url_kwargs_dict_for_review(review))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        reviews = simplejson.loads(response.content)
        self.assertEqual(reviews[0]["app_name"], review.app_name)
        self.assertEqual(reviews[0]["package_name"], review.package_name())

        # now repeat with blender that contains a ()
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            app_name='Blender (windowed)')
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=_get_url_kwargs_dict_for_review(review))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        reviews = simplejson.loads(response.content)
        self.assertEqual(reviews[0]["app_name"], review.app_name)
        self.assertEqual(reviews[0]["package_name"], review.package_name())

        # now repeat with a fake app that contains a ";"
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo2'),
            app_name=u'TestApp with;!"§$%&()=?ß/äluü')
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=_get_url_kwargs_dict_for_review(review))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        reviews = simplejson.loads(response.content)
        self.assertEqual(reviews[0]["app_name"], review.app_name)
        self.assertEqual(reviews[0]["package_name"], review.package_name())

    def test_review_for_language(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=package,
            language='en')
        review2 = self.factory.makeReview(software_item=package,
            language='de')
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'en', 'any', 'any', 'any',
            'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertIn(review1, reviews)
        self.assertNotIn(review2, reviews)

    def test_review_for_any_language(self):
        package = self.factory.makeSoftwareItem('package_foo')
        repository = self.factory.makeRepository('ubuntu', 'hardy')
        review1 = self.factory.makeReview(software_item=package,
            language='en', repository=repository, version='1.0')
        review2 = self.factory.makeReview(software_item=package,
            language='de', repository=repository, version='1.0')
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'any', 'ubuntu', 'hardy', '1.0',
            'package_foo')

        self.assertEqual(2, len(reviews))
        self.assertIn(review1, reviews)
        self.assertIn(review2, reviews)

    def test_review_from_distroseries(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'hardy'))
        review2 = self.factory.makeReview(software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'lucid'))
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'any', 'any', 'hardy', 'any',
            'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertIn(review1, reviews)
        self.assertNotIn(review2, reviews)

    def test_review_from_any_distroseries(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='en', version='+-1')
        review2 = self.factory.makeReview(software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'lucid'),
            language='en', version='+-1')
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'en', 'ubuntu', 'any', '+-1',
            'package_foo')

        self.assertEqual(2, len(reviews))
        self.assertIn(review1, reviews)
        self.assertIn(review2, reviews)

    def test_review_from_origin(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'hardy'))
        review2 = self.factory.makeReview(software_item=package,
            repository=self.factory.makeRepository('debian', 'lenny'))
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'any', 'ubuntu', 'any', 'any',
            'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertIn(review1, reviews)
        self.assertNotIn(review2, reviews)

    def test_review_from_any_origin(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=package,
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='en', version='vanilla')
        review2 = self.factory.makeReview(software_item=package,
            repository=self.factory.makeRepository('debian', 'hardy'),
            language='en', version='vanilla')
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'en', 'any', 'hardy', 'vanilla',
            'package_foo')

        self.assertEqual(2, len(reviews))
        self.assertIn(review1, reviews)
        self.assertIn(review2, reviews)

    def test_review_for_version(self):
        package = self.factory.makeSoftwareItem('package_foo')
        review1 = self.factory.makeReview(software_item=package,
            version='0.2')
        review2 = self.factory.makeReview(software_item=package,
            version='0.3')
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'any', 'any', 'any', '0.2',
            'package_foo')

        self.assertEqual(1, len(reviews))
        self.assertIn(review1, reviews)
        self.assertNotIn(review2, reviews)

    def test_review_for_any_version(self):
        package = self.factory.makeSoftwareItem('package_foo')
        repository = self.factory.makeRepository('ubuntu', 'hardy')
        review1 = self.factory.makeReview(software_item=package,
            language='en', repository=repository, version='1.0')
        review2 = self.factory.makeReview(software_item=package,
            language='en', repository=repository, version='2.0')
        handler = ShowReviewsHandler()

        reviews = handler.read(None, 'en', 'ubuntu', 'hardy', 'any',
            'package_foo')

        self.assertEqual(2, len(reviews))
        self.assertIn(review1, reviews)
        self.assertIn(review2, reviews)

    def test_review_sorting(self):
        package = self.factory.makeSoftwareItem('pkg_foo')
        review1 = self.factory.makeReview(software_item=package)
        review2 = self.factory.makeReview(software_item=package)
        showHandler = ShowReviewsHandler()
        usefulnessHandler = SubmitUsefulnessHandler()

        # 1 useful vote for review2
        request = Mock()
        request.user = self.factory.makeUser()
        usefulnessHandler.create(request, review2.id)
        reviews = showHandler.read(None, 'any', 'any', 'any', 'any',
            'pkg_foo')

        self.assertEqual(2, len(reviews))
        # review2 is sorted first
        self.assertEquals(review2, reviews[0])

        # 2 useful votes for review1
        usefulnessHandler.create(request, review1.id)
        request.user = self.factory.makeUser()
        usefulnessHandler.create(request, review1.id)

        reviews = showHandler.read(None, 'any', 'any', 'any', 'any',
            'pkg_foo')

        self.assertEqual(2, len(reviews))
        # review1 is sorted first
        self.assertEquals(review1, reviews[0])

    def _make_5_reviews_for_package(self, package_name):
        package = self.factory.makeSoftwareItem(package_name)
        for count in range(5):
            self.factory.makeReview(software_item=package)

    def _reviews_url_for_package(self, package_name, origin='any',
                                 distro_series='any', language='any',
                                 version='any',
                                 page=None):
        kwargs = {
            'language': language,
            'origin': origin,
            'distroseries': distro_series,
            'package_name': package_name,
            'version': version,
            }
        if page is not None:
            kwargs['page'] = page

        return reverse('rnr-api-reviews-for-package', kwargs=kwargs)

    def test_only_first_pagination_returned(self):
        # The default request will return only the first batch of
        # reviews.
        self._make_5_reviews_for_package('package_foo')

        with patch_settings(API_REVIEWS_PAGINATION=3):
            response = self.client.get(self._reviews_url_for_package(
                'package_foo'))

        response_decoded = simplejson.loads(response.content)
        self.assertEqual(3, len(response_decoded))

    def test_request_second_page(self):
        # Subsequent pages of results are returned when requested.
        self._make_5_reviews_for_package('package_foo')

        with patch_settings(API_REVIEWS_PAGINATION=3):
            response = self.client.get(self._reviews_url_for_package(
                'package_foo', page=2))

        response_decoded = simplejson.loads(response.content)
        self.assertEqual(2, len(response_decoded))

    def test_request_cached(self):
        # Requests for reviews are cached.
        url = self._reviews_url_for_package('package_foo')
        response = self.client.get(url)

        self.assertEqual(
            'max-age={0}'.format(settings.CACHE_MIDDLEWARE_SECONDS),
            response['cache-control'])
        key = cache_key_for_url(url)
        self.assertTrue(cache.has_key(key))

        # A subsequent call will not hit the view.
        with patch('reviewsapp.api.handlers.ShowReviewsHandler.read') as (
            mock_read):
            response = self.client.get(
                self._reviews_url_for_package('package_foo'))
            self.assertFalse(mock_read.called)

    def test_review_from_id(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_bar'))
        url = reverse(
            'rnr-api-review-for-review-id',
            kwargs={ 'review_id' : review.id })
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        result_review = simplejson.loads(response.content)
        self.assertEqual(result_review["package_name"], review.package_name())
        self.assertEqual(result_review["id"], review.id)


class SubmitReviewHandlerTestCase(TestCaseWithFactory):
    def setUp(self):
        self.required_data = {
            'package_name': 'inkscape',
            'summary': 'Excellent',
            'version': '0.47.0-2ubuntu2',
            'review_text': 'Plain brilliant piece of software',
            'rating': 5,
            'distroseries': 'lucid',
            'origin': 'ubuntu',
            'language': 'en',
            'arch_tag': 'i386',
        }
        super(SubmitReviewHandlerTestCase, self).setUp()

    def _post_new_review(self, data, user=None, verify_result=True):
        # Post a review as an authenticated user.
        url = reverse('rnr-api-reviews')

        if user is None:
            user = self.factory.makeUser()
        self.client.login(username=user.username, password='test')

        # Ensure we don't hit the network for our tests.
        is_authed_fn = 'reviewsapp.auth.SSOOAuthAuthentication.is_authenticated'
        lp_verify_fn = ('reviewsapp.utilities.WebServices.'
                        'lp_verify_packagename_in_repository')
        sca_verify_fn = ('reviewsapp.utilities.WebServices.'
                         'sca_verify_packagename_in_repository')
        with patch(is_authed_fn) as mock_is_authenticated:
            mock_is_authenticated.return_value = True
            with patch(lp_verify_fn) as mock_lp_verify_method:
                mock_lp_verify_method.return_value = verify_result
                with patch(sca_verify_fn) as mock_sca_verify_method:
                    mock_sca_verify_method.return_value = verify_result
                    response = self.client.post(
                        url, data=data, content_type='application/json')
        return response

    def test_bogus_data(self):
        # Submitting bogus data results in a bad request.
        response = self._post_new_review(simplejson.dumps({'foo': 'bar'}))

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_missing_args(self):
        # If required args are not included it is a bad request.
        for key in self.required_data:
            data = self.required_data.copy()
            del data[key]

            response = self._post_new_review(simplejson.dumps(data))

            self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_create(self):
        # Posting the required json data creates and returns the review.
        response = self._post_new_review(simplejson.dumps(self.required_data))

        self.assertEqual(httplib.OK, response.status_code)
        self.assertEqual('application/json; charset=utf-8',
                         response['content-type'])
        response_dict = simplejson.loads(response.content)
        self.assertEqual('inkscape', response_dict['package_name'])

    def test_creates_repository_and_software_item(self):
        # Submitting a review for a software item or repository of which
        # we don't yet have a record will create those records (after
        # validating that the corresponding binary exists on Launchpad).
        required_data = self.required_data
        required_data['distroseries'] = 'doesntexist'
        self.assertEqual(0, Repository.objects.filter(
            origin='ubuntu', distroseries='doesntexist').count())
        self.assertEqual(0, SoftwareItem.objects.filter(
            package_name='inkscape').count())

        response = self._post_new_review(simplejson.dumps(self.required_data))

        self.assertEqual(1, Repository.objects.filter(
            origin='ubuntu', distroseries='doesntexist').count())
        self.assertEqual(1, SoftwareItem.objects.filter(
            package_name='inkscape').count())

    def test_bad_repository(self):
        # If lp_verify returns false a bad request status is returned,
        # and no new repositories are created.
        required_data = self.required_data
        required_data['distroseries'] = 'doesntexist'
        self.assertEqual(0, Repository.objects.filter(
            origin='ubuntu', distroseries='doesntexist').count())
        self.assertEqual(0, SoftwareItem.objects.filter(
            package_name='inkscape').count())

        response = self._post_new_review(
            simplejson.dumps(self.required_data), verify_result=False)

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)
        self.assertEqual(0, Repository.objects.filter(
            origin='ubuntu', distroseries='doesntexist').count())
        self.assertEqual(0, SoftwareItem.objects.filter(
            package_name='inkscape').count())

    def test_max_length_summary(self):
        # Posting with a summary > 80 chars results in a bad request.
        data = self.required_data
        data['summary'] = 'a' * 82

        response = self._post_new_review(simplejson.dumps(data))

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)
        errors = simplejson.loads(response.content)['errors']
        self.assertEqual(['summary'], errors.keys())

    def test_max_length_text(self):
        # Posting with a review_text > 5000 chars results in a bad
        # request.
        data = self.required_data
        data['review_text'] = 'a' * 5027

        response = self._post_new_review(simplejson.dumps(data))

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)
        errors = simplejson.loads(response.content)['errors']
        self.assertEqual(['review_text'], errors.keys())

    def test_stats_updated(self):
        # Posting a review updates the associated stats.
        software_item = self.factory.makeSoftwareItem(
            package_name=self.required_data['package_name'],
            ratings_total=0, ratings_average=0)

        self._post_new_review(simplejson.dumps(self.required_data))

        software_item = SoftwareItem.objects.get(id=software_item.id)
        self.assertEqual(1, software_item.ratings_total)
        self.assertEqual(5, software_item.ratings_average)

    def test_unknown_architecture(self):
        # Posting a review of an unknown architecture is a bad request.
        self.required_data['arch_tag'] = 'unknown'

        response = self._post_new_review(simplejson.dumps(self.required_data))

        errors = simplejson.loads(response.content)['errors']
        self.assertEqual(1, len(errors))
        self.assertTrue(errors.has_key('arch_tag'))

    def test_multiple_reviews_per_arch(self):
        # Posting multiple reviews for the same architecture shouldn't fail
        for i in range(3):
            response = self._post_new_review(
                simplejson.dumps(self.required_data))
            self.assertEqual(200, response.status_code)

    def test_submit_review_invalidates_cache(self):
        # Submitting a new review invalidates the show reviews cache for
        # the package.
        software_item = self.factory.makeSoftwareItem(
            package_name=self.required_data['package_name'],
            ratings_total=0, ratings_average=0)
        url_keys = ('language', 'origin', 'distroseries', 'version',
            'package_name')
        url_kwargs = dict((key, self.required_data[key]) for key in url_keys)
        url = reverse('rnr-api-reviews-for-package', kwargs=url_kwargs)
        self.client.get(url)
        key = cache_key_for_url(url)
        self.assertTrue(cache.has_key(key))

        response = self._post_new_review(simplejson.dumps(self.required_data))

        self.assertFalse(cache.has_key(key))

    def test_submit_same_app_and_distroseries_twice_fails(self):
        # A user can only submit a review for an app once.
        user = self.factory.makeUser()
        first = self._post_new_review(simplejson.dumps(self.required_data),
            user=user)
        second = self._post_new_review(simplejson.dumps(self.required_data),
            user=user)

        self.assertEqual(httplib.OK, first.status_code)
        self.assertEqual(httplib.BAD_REQUEST, second.status_code)
        errors = simplejson.loads(second.content)['errors']
        self.assertEqual(['__all__'], errors.keys())

    def test_submit_same_app_differente_series_ok(self):
        # A user can submit multiple reviews for the same app in
        # different distroseries.
        user = self.factory.makeUser()
        first = self._post_new_review(simplejson.dumps(self.required_data),
            user=user)
        data_natty = self.required_data
        data_natty.update(distroseries='natty')
        second = self._post_new_review(simplejson.dumps(data_natty),
            user=user)

        self.assertEqual(httplib.OK, first.status_code)
        self.assertEqual(httplib.OK, second.status_code)


class FlagReviewHandlerTestCase(TestCaseWithFactory):
    def _flag_review_id(self, review_id, user=None, data=None):
        if user is None:
            user = self.factory.makeUser()
        if data is None:
            data = dict(reason='A reason', text='explanation of the flag')
        request = Mock()
        request.POST = data
        request.user = user
        handler = FlagReviewHandler()
        return handler.create(request, review_id)

    def test_bogus_data(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package_foo'),
            app_name='app_foo',
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='en')

        response = self._flag_review_id(review.id, data={'foo': 'bazzinga'})

        self.assertEqual(httplib.BAD_REQUEST, response.status_code)

    def test_bad_review_id(self):
        response = self._flag_review_id(1234)

        self.assertEqual(httplib.NOT_FOUND, response.status_code)

    def test_create(self):
        review = self.factory.makeReview(
            software_item=self.factory.makeSoftwareItem('package-foo'),
            app_name=u'Déjà Dup',
            repository=self.factory.makeRepository('ubuntu', 'hardy'),
            language='en')
        flagger = self.factory.makeUser()

        flag = self._flag_review_id(review.id, user=flagger)

        self.assertTrue(isinstance(flag, ReviewModerationFlag))
        self.assertEqual(flagger, flag.user)
        self.assertEqual(review, flag.review_moderation.review)

    def test_flag_review_invalidates_cache(self):
        # Flagging a new review invalidates the show reviews cache for
        # the package only if the review visibility changes.
        review = self.factory.makeReview()
        url_kwargs = dict(language=review.language,
            origin=review.repository.origin,
            distroseries=review.repository.distroseries,
            package_name=review.softwareitem.package_name,
            version=review.version)
        url = reverse('rnr-api-reviews-for-package', kwargs=url_kwargs)
        self.client.get(url)
        key = cache_key_for_url(url)
        self.assertTrue(cache.has_key(key))

        self._flag_review_id(review.id)

        self.assertFalse(cache.has_key(key))


class UsefulnessHandlerTestCase(TestCaseWithFactory):

    def test_vote_on_usefulness(self):
        review = self.factory.makeReview()
        self.assertEqual(review.usefulness_total, 0)
        self.assertEqual(review.usefulness_favorable, 0)
        self.assertEqual(review.usefulness_percentage, 0)
        request = Mock()
        request.user = self.factory.makeUser()

        response = SubmitUsefulnessHandler().create(request, review.id)

        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 1)
        self.assertEqual(review.usefulness_percentage, 100)

    def test_usefulness_in_review(self):
        review = self.factory.makeReview()
        request = Mock()
        request.user = self.factory.makeUser()
        handler = SubmitUsefulnessHandler()
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        url = reverse(
            'rnr-api-reviews-for-package',
            kwargs=_get_url_kwargs_dict_for_review(review))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        reviews = simplejson.loads(response.content)
        self.assertEqual(reviews[0]["usefulness_total"], 1)
        self.assertEqual(reviews[0]["usefulness_favorable"], 1)

    def test_one_vote_per_user(self):
        review = self.factory.makeReview()
        request = Mock()
        request.user = self.factory.makeUser()
        handler = SubmitUsefulnessHandler()
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 1)
        self.assertEqual(review.usefulness_percentage, 100)
        # same user
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(response.content, '"Not modified"')
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 1)
        self.assertEqual(review.usefulness_percentage, 100)
        # same user
        request.POST = {'useful': 'False'}
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(response.content, '"Updated"')
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 0)
        self.assertEqual(review.usefulness_percentage, 0)

    def test_vote_aggregation(self):
        review = self.factory.makeReview()
        request = Mock()
        request.user = self.factory.makeUser()
        handler = SubmitUsefulnessHandler()
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 1)
        self.assertEqual(review.usefulness_favorable, 1)
        self.assertEqual(review.usefulness_percentage, 100)
        # new user
        request.user = self.factory.makeUser()
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 2)
        self.assertEqual(review.usefulness_favorable, 2)
        self.assertEqual(review.usefulness_percentage, 100)
        # another new user
        request.user = self.factory.makeUser()
        request.POST = {'useful': 'False'}
        response = handler.create(request, review.id)
        self.assertEquals(response.status_code, 201)
        review = Review.objects.get(id=review.id)
        self.assertEqual(review.usefulness_total, 3)
        self.assertEqual(review.usefulness_favorable, 2)
        self.assertEqual(review.usefulness_percentage, 66)

    def _post_usefulness_as_user(self, user, useful=True):
        review = self.factory.makeReview()
        url = reverse('rnr-api-recommendations', args=[review.id])
        self.client.login(username=user.username, password='test')
        is_authenticated_method = (
            'reviewsapp.auth.SSOOAuthAuthentication.is_authenticated')
        with patch(is_authenticated_method) as mock_is_authenticated:
            mock_is_authenticated.return_value = True
            response = self.client.post(url, data={'useful': useful})
        return response

    def test_request_not_cached(self):
        # Requests for server status are never cached.
        user = self.factory.makeUser()

        response = self._post_usefulness_as_user(user)

        self.assertFalse(response.has_header('cache-control'))

        # A subsequent call will hit the view.
        create_method = (
            'reviewsapp.api.handlers.SubmitUsefulnessHandler.create')
        with patch(create_method) as mock_create:
            response = self._post_usefulness_as_user(user)
            self.assertTrue(mock_create.called)
