#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for our middleware classes"""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'StatsLoggingTestCase',
    ]

import logging
import time

from django.core.urlresolvers import reverse
from django.test import TestCase
from mock import patch, MagicMock, Mock

from reviewsapp.middleware.stats import StatisticsLoggingMiddleware
from reviewsapp.middleware.exception import LogExceptionMiddleware
from reviewsapp.tests.factory import TestCaseWithFactory
from reviewsapp.tests.helpers import patch_settings

class StatsLoggingTestCase(TestCaseWithFactory):
    @patch('logging.Logger.info')
    def test_enabled(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           STATS_TO_LOG=[StatisticsLoggingMiddleware.LOG_CPU]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertTrue(mock_info.called)

    @patch('logging.Logger.info')
    def test_disabled(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=False,
                            SECONDS_BETWEEN_LOGGING=1,
                           STATS_TO_LOG=[StatisticsLoggingMiddleware.LOG_CPU]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertFalse(mock_info.called)

    @patch('logging.Logger.info')
    def test_time_slice_met(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           STATS_TO_LOG=[StatisticsLoggingMiddleware.LOG_CPU]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertTrue(mock_info.called)

    @patch('logging.Logger.info')
    def test_time_slice_not_met(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=3,
                           STATS_TO_LOG=[StatisticsLoggingMiddleware.LOG_CPU]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertFalse(mock_info.called)

    @patch('logging.Logger.info')
    def test_valid_device(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           DISKS=['sda13'],
                           STATS_TO_LOG=[StatisticsLoggingMiddleware.LOG_DISK]):

            StatisticsLoggingMiddleware.last_run = time.time() - 2
            open_name = 'linux_metrics.disk_stat.open'
            with patch(open_name, create=True) as mock_open:
                mock_file = MagicMock(spec=file, name='MockFile')
                mock_file.read.return_value = '1       0 sda13 52571 48009 2091454 1474440 62473 233806 2365864 2223730 0 508610 3699930\n'
                enter_mock = Mock()
                exit_mock = Mock()
                context_mgr = Mock()
                mock_open.return_value = context_mgr
                enter_mock.return_value = mock_file
                setattr(context_mgr, '__enter__', enter_mock)
                setattr(context_mgr, '__exit__', exit_mock)

                self.client.get(reverse('rnr-overview'))
            self.assertTrue(mock_info.called)
            logged = mock_info.call_args
            self.assertTrue('DISK (sda13)' in logged[0][0])


    @patch('logging.Logger.info')
    def test_invalid_device(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           DISKS=['bogus'],
                           STATS_TO_LOG=[StatisticsLoggingMiddleware.LOG_DISK]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertTrue(mock_info.called)
            logged = mock_info.call_args
            self.assertTrue('error' in logged[0][0])

    @patch('logging.Logger.info')
    def test_invalid_stats_to_log_setting(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           STATS_TO_LOG=['bogus']):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertFalse(mock_info.called)

    @patch('logging.Logger.info')
    def test_log_cpu_only(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           STATS_TO_LOG=[StatisticsLoggingMiddleware.LOG_CPU]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertTrue(mock_info.called)
            logged = mock_info.call_args[0]
            self.assertFalse('DISK' in logged[0])
            self.assertFalse('NET' in logged[0])
            self.assertFalse('MEM' in logged[0])
            self.assertTrue('CPU' in logged[0])

    @patch('logging.Logger.info')
    def test_log_cpu_and_mem_only(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           STATS_TO_LOG=[
                               StatisticsLoggingMiddleware.LOG_CPU,
                               StatisticsLoggingMiddleware.LOG_MEMORY]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertTrue(mock_info.called)
            logged = mock_info.call_args[0]
            self.assertFalse('DISK' in logged[0])
            self.assertFalse('NET' in logged[0])
            self.assertTrue('MEM' in logged[0])
            self.assertTrue('CPU' in logged[0])

    @patch('logging.Logger.info')
    def test_log_mem_only(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           STATS_TO_LOG=[StatisticsLoggingMiddleware.LOG_MEMORY]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertTrue(mock_info.called)
            logged = mock_info.call_args[0]
            self.assertFalse('DISK' in logged[0])
            self.assertFalse('NET' in logged[0])
            self.assertTrue('MEM' in logged[0])
            self.assertFalse('CPU' in logged[0])

    @patch('logging.Logger.info')
    def test_log_net_only(self, mock_info):
        with patch_settings(ENABLE_STATISTICS_LOGGING=True,
                            SECONDS_BETWEEN_LOGGING=1,
                           STATS_TO_LOG=[
                               StatisticsLoggingMiddleware.LOG_NETWORK]):
            StatisticsLoggingMiddleware.last_run = time.time() - 2
            self.client.get(reverse('rnr-overview'))
            self.assertTrue(mock_info.called)
            logged = mock_info.call_args[0]
            self.assertFalse('DISK' in logged[0])
            self.assertTrue('NET' in logged[0])
            self.assertFalse('MEM' in logged[0])
            self.assertFalse('CPU' in logged[0])


class ExceptionMiddlewareTestCase(TestCase):

    POST_PASSWORD = 'P1!2@3#'
    POST_SECRET = 'P2@3#4$'
    POST_PRIVATE = 'P3#4$5%'
    GET_PASSWORD = 'G1!2@3#'
    GET_SECRET = 'G2@3#4$'
    GET_PRIVATE = 'G3#4$5%'
    POST_PUBLIC = 'Public Post Data'
    GET_PUBLIC = 'Public Post Data'
    asterisks = '********'
    PRIVATE_SETTING = '7&8*9('
    PUBLIC_SETTING = 'Something public'
    FRAME_VAR = '*8*8*8*8'

    LOG_FILENAME = '/tmp/logging_example.out'
    logging.basicConfig(filename=LOG_FILENAME, level=logging.DEBUG)

    logging.debug('This message should go to the log file')

    def setUp(self):
        request = MockRequest('/login')
        request.POST = {}
        request.GET = {}
        request.COOKIES = {}
        request.META = {'SERVER_NAME': 'localhost', 'SERVER_PORT': 80}
        request.POST['post_password'] = self.POST_PASSWORD
        request.POST['post_secret'] = self.POST_SECRET
        request.POST['post_private'] = self.POST_PRIVATE
        request.POST['post_public'] = self.POST_PUBLIC
        request.GET['get_password'] = self.GET_PASSWORD
        request.GET['get_secret'] = self.GET_SECRET
        request.GET['get_private'] = self.GET_PRIVATE
        request.GET['get_public'] = self.GET_PUBLIC
        self.request = request
        self.middleware = LogExceptionMiddleware()
        settings.PRIVATE_SETTING = self.PRIVATE_SETTING
        settings.PUBLIC_SETTING = self.PUBLIC_SETTING

    def test_settings_sanitization(self):
        try:
            x = 2 / 0
        except:
            reporter = self.middleware.get_reporter(self.request,
                                                    *sys.exc_info())
            html = reporter.get_traceback_html()
        self.assertFalse(self.PRIVATE_SETTING in html)
        self.assertTrue(self.PUBLIC_SETTING in html)

    def test_frames_sanitization(self):
        try:
            FRAME_PASSWORD = self.FRAME_VAR
            public_variable = 2 / 0
        except:
            reporter = self.middleware.get_reporter(self.request,
                                                    *sys.exc_info())
            html = reporter.get_traceback_html()
        self.assertFalse(self.FRAME_VAR in html)
        self.assertTrue('public_variable' in html)

    def test_request_sanitization(self):
        req = self.middleware.sanitize_request(self.request)
        #  Make sure sensitive data is no longer in the request
        self.assertFalse(self.POST_PASSWORD in req.POST.values())
        self.assertFalse(self.POST_SECRET in req.POST.values())
        self.assertFalse(self.POST_PRIVATE in req.POST.values())
        self.assertFalse(self.GET_PASSWORD in req.GET.values())
        self.assertFalse(self.GET_SECRET in req.GET.values())
        self.assertFalse(self.GET_PRIVATE in req.GET.values())
        #  Make sure sensitive data is replaced by asterisks
        self.assertEquals(req.POST['post_password'], self.asterisks)
        self.assertEquals(req.POST['post_secret'], self.asterisks)
        self.assertEquals(req.POST['post_private'], self.asterisks)
        self.assertEquals(req.GET['get_password'], self.asterisks)
        self.assertEquals(req.GET['get_secret'], self.asterisks)
        self.assertEquals(req.GET['get_private'], self.asterisks)
        #  Make sure public data is unaffected
        self.assertEquals(req.POST['post_public'], self.POST_PUBLIC)
        self.assertEquals(req.GET['get_public'], self.GET_PUBLIC)

