#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the RnR api client module."""

from __future__ import (
    absolute_import,
    with_statement,
    )

__metaclass__ = type
__all__ = [
    'RnRFlagReviewTestCase',
    'RnRGetReviewTestCase',
    'RnRGetReviewsTestCase',
    'RnRReviewStatsTestCase',
    'RnRServerStatusTestCase',
    'RnRSubmitReviewTestCase',
    'UsefulnessAPITestCase',
    ]

from datetime import timedelta

from decimal import Decimal
from django.core.handlers.wsgi import WSGIHandler
from mock import (
    Mock,
    patch,
    )
from piston_mini_client.validators import ValidationException
from piston_mini_client.auth import OAuthAuthorizer
from rnrclient import RatingsAndReviewsAPI, ReviewRequest

from reviewsapp.models import (
    ReviewModeration,
    ReviewModerationFlag,
    RNRSettings,
    Usefulness,
    )
from reviewsapp.tests.helpers import (
    patch_settings,
    WSGIInterceptedTestCase,
    mock_data_for_account,
    )
from reviewsapp.tests.factory import (
    TestCaseWithFactory,
    TxTestCaseWithFactory,
    )
from reviewsapp.utilities import invalidate_paginated_reviews_for


class RnRBaseTestCase(WSGIInterceptedTestCase, TestCaseWithFactory):
    def setUp(self):
        callbacks = {('localhost', 8000): WSGIHandler}
        super(RnRBaseTestCase, self).setUp(callbacks)


class RnRTxBaseTestCase(WSGIInterceptedTestCase, TxTestCaseWithFactory):
    def setUp(self):
        callbacks = {('localhost', 8000): WSGIHandler}
        super(RnRTxBaseTestCase, self).setUp(callbacks)


class RnRServerStatusTestCase(RnRBaseTestCase):
    def test_server_status(self):
        api = RatingsAndReviewsAPI()

        response = api.server_status()

        self.assertEqual('ok', response)


class RnRReviewStatsTestCase(RnRTxBaseTestCase):

    def get_stats_without_cache(self, **kwargs):
        # We just ensure that the result be taken from the cache.
        with patch('django.core.cache.cache.get') as mock_cache_get:
            mock_cache_get.return_value = None
            stats = RatingsAndReviewsAPI().review_stats(**kwargs)
        return stats

    def make_app_with_ratings(self, package_name, ratings=[],
        origin='ubuntu'):
        repo = self.factory.makeRepository(origin=origin)
        software_item = self.factory.makeSoftwareItem(
            package_name=package_name)
        for rating in ratings:
            self.factory.makeReview(
                software_item=software_item, rating=rating,
                repository=repo)
        return software_item

    def test_review_stats_single_package(self):
        self.make_app_with_ratings('SuperCoolApp', [3, 4, 5])

        stats = self.get_stats_without_cache()

        self.assertEquals(1, len(stats))  # one per software item
        self.assertEquals(3, stats[0].ratings_total)
        self.assertEquals(Decimal('4'), Decimal(stats[0].ratings_average))

    def test_review_stats_multiple_packages(self):
        self.make_app_with_ratings('SuperCoolApp', [3, 4, 5])
        self.make_app_with_ratings('SortaCoolApp', [1, 2, 3])

        stats = self.get_stats_without_cache()

        self.assertEquals(2, len(stats))  # one per software item
        self.assertEquals(3, stats[1].ratings_total)
        self.assertEquals(Decimal('2'), Decimal(stats[1].ratings_average))

        # Check stats for pkg 1 are still the same
        self.assertEquals(3, stats[0].ratings_total)
        self.assertEquals(Decimal('4'), Decimal(stats[0].ratings_average))

    def test_review_stats_updates_only(self):
        # Apps with stats updated in the last N days only are returned.
        self.make_app_with_ratings('SuperCoolApp', [3, 4, 5])
        sorta_cool_app = self.make_app_with_ratings('SortaCoolApp', [1, 2, 3])
        # SortaCoolApp had its ratings last updated 5 days ago.
        sorta_cool_app.date_ratings_updated = (
            sorta_cool_app.date_ratings_updated - timedelta(5))
        sorta_cool_app.save()

        self.assertEquals(2, len(self.get_stats_without_cache()))
        self.assertEquals(1, len(self.get_stats_without_cache(days=1)))
        self.assertEquals(1, len(self.get_stats_without_cache(days=3)))
        self.assertEquals(2, len(self.get_stats_without_cache(days=7)))

    def test_review_stats_has_no_vary_header(self):
        """Check that no Vary header is sent out, as it spoils caching."""
        response = self.client.get('/reviews/api/1.0/review-stats/')
        self.assertFalse(response.has_header('vary'))

    def test_review_stats_ppa_origin(self):
        self.make_app_with_ratings('test-package', [3, 4, 5],
            origin='lp-ppa-app-review-board')

        stats = self.get_stats_without_cache(origin='lp-ppa-app-review-board')

        self.assertEquals(1, len(stats))


class RnRGetReviewsTestCase(RnRBaseTestCase):
    def test_get_reviews_for_package(self):
        review = self.factory.makeReview()
        api = RatingsAndReviewsAPI()

        reviews = api.get_reviews(language=review.language,
            origin=review.repository.origin,
            distroseries=review.repository.distroseries,
            packagename=review.package_name()
        )

        self.assertEqual(1, len(reviews))
        self.assertEqual(review.package_name(), reviews[0].package_name)
        self.assertEqual(review.app_name, reviews[0].app_name)
        self.assertEqual(review.summary, reviews[0].summary)
        self.assertEqual(review.version, reviews[0].version)
        self.assertEqual(review.reviewer_username(),
                         reviews[0].reviewer_username)
        self.assertEqual(review.reviewer_displayname(),
                         reviews[0].reviewer_displayname)
        self.assertEqual(review.origin(), reviews[0].origin)
        self.assertEqual(review.distroseries(), reviews[0].distroseries)

    def test_get_reviews_for_extras_pkg(self):
        repo = self.factory.makeRepository(origin='lp-ppa-app-review-board')
        review = self.factory.makeReview(repository=repo)
        api = RatingsAndReviewsAPI()

        reviews = api.get_reviews(language=review.language,
            origin=review.repository.origin,
            distroseries=review.repository.distroseries,
            packagename=review.package_name()
        )

        self.assertEqual(1, len(reviews))

    def test_get_reviews_for_app(self):
        review = self.factory.makeReview(app_name='something')
        api = RatingsAndReviewsAPI()

        reviews = api.get_reviews(language=review.language,
            origin=review.repository.origin,
            distroseries=review.repository.distroseries,
            packagename=review.package_name(),
            appname=review.app_name,
            version=review.version
        )

        self.assertEqual(1, len(reviews))
        self.assertEqual(review.package_name(), reviews[0].package_name)
        self.assertEqual(review.app_name, reviews[0].app_name)
        self.assertEqual(review.summary, reviews[0].summary)
        self.assertEqual(review.version, reviews[0].version)

    def test_get_reviews_page_2(self):
        # The rnr_client can get multiple pages of results.
        software_item = self.factory.makeSoftwareItem()
        repository = self.factory.makeRepository()
        for count in range(3):
            self.factory.makeReview(software_item=software_item,
                language='en', repository=repository)

        with patch_settings(API_REVIEWS_PAGINATION=2):
            reviews = RatingsAndReviewsAPI().get_reviews(language='en',
                origin=repository.origin,
                distroseries=repository.distroseries,
                packagename=software_item.package_name, page=2)

        self.assertEqual(1, len(reviews))

    def test_get_reviews_has_no_vary_header(self):
        """Check that no Vary header is sent out, as it spoils caching."""
        response = self.client.get(
            '/reviews/api/1.0/reviews/filter/en/ubuntu/kinky/any/chess/')

        self.assertEqual(200, response.status_code)
        self.assertFalse(response.has_header('vary'))


class RnRGetReviewTestCase(RnRBaseTestCase):

    def test_get_review(self):
        review = self.factory.makeReview()

        api = RatingsAndReviewsAPI()
        review_detail = api.get_review(review_id=int(review.id))

        self.assertEqual(review_detail.id, review.id)
        self.assertEqual(review_detail.origin, review.origin())
        self.assertEqual(review_detail.distroseries, review.distroseries())
        invalidate_paginated_reviews_for(review)


class RnRSubmitReviewTestCase(RnRTxBaseTestCase):

    def _submit_review(self, review_request, token_and_consumer=None,
        mock_verify_package=None):
        """Submit a review request and return the result with a mock."""
        if token_and_consumer is None:
            token, consumer = self.factory.makeOAuthTokenAndConsumer(
                user=self.factory.makeUser())
        else:
            token, consumer = token_and_consumer

        auth = OAuthAuthorizer(token.token, token.token_secret, consumer.key,
            consumer.secret)
        lp_verify_packagename_method = (
            'reviewsapp.utilities.WebServices.'
            'lp_verify_packagename_in_repository')

        if mock_verify_package is None:
            mock_verify_package = Mock()
            mock_verify_package.return_value = True

        api = RatingsAndReviewsAPI(auth=auth)
        with patch(lp_verify_packagename_method, mock_verify_package):
            response = api.submit_review(review=review_request)

        return response, mock_verify_package

    def _make_review_request(self, **kwargs):
        review_options = dict(package_name='pacman', summary='Too difficult',
            version='0.1.2', rating=3, arch_tag='i386',
            review_text='I keep getting mowed down by ghosts',
            language='en', origin='ubuntu', distroseries='lucky')
        review_options.update(kwargs)
        return ReviewRequest(**review_options)

    def test_submit_review(self):
        review = self._make_review_request(package_name=u'myapp',
            rating=4, review_text=u'This is great.')

        review_result, ignored = self._submit_review(review)

        self.assertEqual(u'myapp', review_result.package_name)
        self.assertEqual(4, review_result.rating)
        self.assertEqual(u'This is great.', review_result.review_text)
        self.assertFalse(review_result.hide)

    def test_submit_review_anonymous(self):
        api = RatingsAndReviewsAPI()
        self.assertRaises(ValidationException, api.submit_review)

    @patch('reviewsapp.utilities.WebServices.get_data_for_account')
    def test_submit_review_fetch(self, mock_get_data):
        """Submit a review using a token/consumer that we don't have.

        The auth layer will need to fetch the information from SSO.
        """
        # Create a token/consumer pair without saving locally, and ensure that
        # it is returned by the get_data_for_account.
        token, consumer = self.factory.makeOAuthTokenAndConsumer(
            user=self.factory.makeUser(), save=False)
        mock_get_data.return_value = mock_data_for_account(consumer, token)
        review = self._make_review_request()

        review_result, ignored = self._submit_review(
            review, (token, consumer))

        self.assertEqual('pacman', review_result.package_name)
        self.assertEqual(1, mock_get_data.call_count)

    def test_submit_review_nonexisting_distroseries(self):
        review = self._make_review_request(
            package_name='pacman', distroseries='lucky',
            review_text='First review')

        # First time for this package_name & distroseries - should call verify
        mock_verify_package = Mock()
        response, mock_verify_package = self._submit_review(review,
            mock_verify_package=mock_verify_package)

        self.assertEqual(mock_verify_package.call_count, 1)

        # Same package_name & distroseries - should not call verify
        review = self._make_review_request(
            package_name='pacman', distroseries='lucky',
            review_text='Second review')
        response, mock_verify_package = self._submit_review(review,
            mock_verify_package=mock_verify_package)

        self.assertEqual(mock_verify_package.call_count, 1)

        # Change the distroseries - should call verify
        review = self._make_review_request(
            package_name='pacman', distroseries='bogus',
            review_text='Third review')
        response, mock_verify_package = self._submit_review(review,
            mock_verify_package=mock_verify_package)
        self.assertEqual(mock_verify_package.call_count, 2)

    def test_submit_review_different_archs(self):
        review = self._make_review_request(
            package_name='pacman', arch_tag='i386',
            review_text='First review')

        # First time for this package_name & arch - should call verify
        mock_verify_package = Mock()
        response, mock_verify_package = self._submit_review(review,
            mock_verify_package=mock_verify_package)

        self.assertEqual(mock_verify_package.call_count, 1)

        # Same package_name & arch - should not call verify
        review = self._make_review_request(
            package_name='pacman', arch_tag='i386',
            review_text='Second review')
        response, mock_verify_package = self._submit_review(review,
            mock_verify_package=mock_verify_package)

        self.assertEqual(mock_verify_package.call_count, 1)

        # Change the arch - should call verify
        review = self._make_review_request(
            package_name='pacman', arch_tag='amd64',
            review_text='Third review')
        response, mock_verify_package = self._submit_review(review,
            mock_verify_package=mock_verify_package)
        self.assertEqual(mock_verify_package.call_count, 2)

    def test_passive_mode(self):
        # The server is in passive mode by default
        review = self._make_review_request()

        review_result, ignored = self._submit_review(review)
        # In passive mode the review will not be hidden
        self.assertFalse(review_result.hide)

        # ensure that no ReviewModerationFlag was created
        self.assertEqual(ReviewModeration.objects.count(), 0)
        self.assertEqual(ReviewModerationFlag.objects.count(), 0)

    def test_active_mode(self):
        # Put the server in active mode
        RNRSettings.set_moderation_mode(RNRSettings.ACTIVE_MODERATION)

        review = self._make_review_request()

        review_result, ignored = self._submit_review(review)
        # In active mode the review will be hidden
        self.assertTrue(review_result.hide)

        # ensure that a ReviewModerationFlag was created
        self.assertEquals(ReviewModeration.objects.count(), 1)
        self.assertEquals(ReviewModerationFlag.objects.count(), 1)
        moderation = ReviewModeration.objects.all()[0]
        flag = ReviewModerationFlag.objects.all()[0]
        # ensure that the review is in pending status
        self.assertEqual(moderation.status, ReviewModeration.PENDING_STATUS)


class RnRFlagReviewTestCase(RnRBaseTestCase):

    def _flag_review(self, review_id, reason, text):
        token, consumer = self.factory.makeOAuthTokenAndConsumer(
            user=self.factory.makeUser())

        auth = OAuthAuthorizer(token.token, token.token_secret, consumer.key,
            consumer.secret)

        api = RatingsAndReviewsAPI(auth=auth)
        return api.flag_review(review_id=int(review_id),
            reason=reason, text=text)

    def test_flag_review_returns_new_flag(self):
        # The newly created flag is returned when calling api.flag_review.
        review = self.factory.makeReview()

        flag = self._flag_review(review.id, 'my_reason', 'my text')

        self.assertEqual('my_reason', flag['summary'])


class UsefulnessAPITestCase(RnRTxBaseTestCase):
    def setUp(self):
        super(UsefulnessAPITestCase, self).setUp()
        self.user = self.factory.makeUser()

    def test_submit_useful(self):
        token, consumer = self.factory.makeOAuthTokenAndConsumer(
            user=self.user)
        auth = OAuthAuthorizer(token.token, token.token_secret, consumer.key,
            consumer.secret)
        api = RatingsAndReviewsAPI(auth=auth)
        software_item = self.factory.makeSoftwareItem()
        review = self.factory.makeReview(software_item=software_item)
        response = api.submit_usefulness(review_id=int(review.id),
            useful="True")
        self.assertEqual('Created', response)
        count = Usefulness.objects.filter(review=review, useful=True).count()
        self.assertEqual(1, count)

    def test_submit_not_useful(self):
        token, consumer = self.factory.makeOAuthTokenAndConsumer(
            user=self.user)
        auth = OAuthAuthorizer(token.token, token.token_secret, consumer.key,
            consumer.secret)
        api = RatingsAndReviewsAPI(auth=auth)
        software_item = self.factory.makeSoftwareItem()
        review = self.factory.makeReview(software_item=software_item)
        response = api.submit_usefulness(review_id=int(review.id),
            useful="False")
        self.assertEqual('Created', response)
        count = Usefulness.objects.filter(review=review, useful=True).count()
        self.assertEqual(0, count)

    def test_submit_useful_multiple_times(self):
        token, consumer = self.factory.makeOAuthTokenAndConsumer(
            user=self.user)
        auth = OAuthAuthorizer(token.token, token.token_secret, consumer.key,
            consumer.secret)
        api = RatingsAndReviewsAPI(auth=auth)
        software_item = self.factory.makeSoftwareItem()
        review = self.factory.makeReview(software_item=software_item)
        response = api.submit_usefulness(review_id=int(review.id),
            useful="True")
        self.assertEqual('Created', response)
        count = Usefulness.objects.filter(review=review, useful=True).count()
        self.assertEqual(1, count)
        # submit same again and check that we don't get created
        response = api.submit_usefulness(
            review_id=int(review.id), useful="True")
        self.assertEqual('Not modified', response)
        count = Usefulness.objects.filter(review=review, useful=True).count()
        self.assertEqual(1, count)
        # submit again, this time with different usefulness
        response = api.submit_usefulness(
            review_id=int(review.id), useful="False")
        self.assertEqual('Updated', response)
        count = Usefulness.objects.filter(review=review).count()
        self.assertEqual(1, count)

    def test_get_usefulness(self):
        api = RatingsAndReviewsAPI()
        software_item = self.factory.makeSoftwareItem()
        review = self.factory.makeReview(software_item=software_item)
        user = self.factory.makeUser()
        usefulness = self.factory.makeUsefulness(review=review, user=user)

        response = api.get_usefulness(review_id=int(review.id),
            username=user.username)

        self.assertEqual(1, len(response))
        self.assertEqual(user.username, response[0].get('username'))
        self.assertEqual(review.id, response[0].get('review_id'))
        self.assertEqual(usefulness.useful, response[0].get('useful'))


