#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the web services utility """

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'FullClaimedIdTestCase',
    'IdentityProviderTestCase',
    'InvalidatePaginatedReviewsTestCase',
    'LaunchpadTestCase',
    'SCATestCase',
    ]

import os
import urllib

from django.conf import settings
from django.core.cache import cache
from django.core.urlresolvers import reverse
from launchpadlib.errors import HTTPError
from mock import patch, Mock
from mockssoservice.mockserver import MockSSOServer, new_token
from wsgi_intercept import add_wsgi_intercept, remove_wsgi_intercept
from wsgi_intercept.httplib2_intercept import install, uninstall

from reviewsapp.tests.helpers import patch_settings
from reviewsapp.tests.factory import TestCaseWithFactory
from reviewsapp.utilities import (
    cache_key_for_url,
    invalidate_paginated_reviews_for,
    WebServiceError,
    WebServices,
    full_claimed_id,
)


class IdentityProviderTestCase(TestCaseWithFactory):
    def mock_sso(self, *args):
        return MockSSOServer('login.ubuntu.com', 443, '/api/1.0',
                             scheme='https')

    def setUp(self):
        install()
        add_wsgi_intercept('login.ubuntu.com', 443, self.mock_sso)
        super(IdentityProviderTestCase, self).setUp()

        self.orig_sso_api_service_root = settings.SSO_API_SERVICE_ROOT
        self.orig_sso_api_auth_username = settings.SSO_API_AUTH_USERNAME
        self.orig_sso_api_auth_password = settings.SSO_API_AUTH_PASSWORD

        settings.SSO_API_SERVICE_ROOT = 'https://login.ubuntu.com/api/1.0'
        # Use the credentials for the mockserver
        settings.SSO_API_AUTH_USERNAME = 'MyUsername'
        settings.SSO_API_AUTH_PASSWORD = 'password'

        # Force our webservice client to recreate the the ServiceRoot
        WebServices._identity_provider = None

    def tearDown(self):
        super(IdentityProviderTestCase, self).tearDown()
        remove_wsgi_intercept('login.ubuntu.com', 443)
        uninstall()
        settings.SSO_API_SERVICE_ROOT = self.orig_sso_api_service_root
        settings.SSO_API_AUTH_USERNAME = self.orig_sso_api_auth_username
        settings.SSO_API_AUTH_PASSWORD = self.orig_sso_api_auth_password

    def test_identity_provider_unavailable(self):
        settings.SSO_API_SERVICE_ROOT = 'http://nothing.localhost/nonapi/3.3'

        self.assertRaises(
            WebServiceError, getattr, WebServices(), 'identity_provider')

    def test_get_data_for_account(self):
        # We tell the mockserver about the token we'll be checking.
        new_token('test-token', 'eg_token')
        web_services = WebServices()
        data = web_services.get_data_for_account('eg_token', 'eg_ident')
        self.assertEquals(set([u'username', u'preferred_email',
            u'displayname', u'name', u'unverified_emails', u'verified_emails',
            u'consumer_secret', u'token', u'openid_identifier',
            u'consumer_key', u'token_secret']), set(data))


class LaunchpadTestCase(TestCaseWithFactory):
    def setUp(self):
        super(LaunchpadTestCase, self).setUp()
        # Force web_services to recreate the the Launchpad service root
        WebServices._launchpad_service = None

    def _make_mock_launchpad(self, binary_name='mypkg'):
        mock_ubuntu = Mock()
        mock_ubuntu.self_link = 'http://lpapi/ubuntu'

        mock_launchpad = Mock()
        mock_launchpad.distributions =  dict(ubuntu=mock_ubuntu)
        app_review_board = Mock()
        mock_launchpad.people = {'app-review-board': app_review_board}

        mock_archive = Mock()
        mock_ubuntu.getArchive.return_value = mock_archive
        app_review_board.getPPAByName.return_value = mock_archive

        mock_binary = Mock()
        mock_binary.binary_package_name = binary_name
        mock_archive.getPublishedBinaries.return_value = [mock_binary]
        return (mock_launchpad, mock_ubuntu, mock_archive)

    def _verify_package(self, mock_lp_api, package_name,
        distro, series, arch_tag):
        with patch('reviewsapp.utilities.Launchpad') as mock_launchpad:
            mock_launchpad.login_anonymously.return_value = mock_lp_api
            return WebServices().lp_verify_packagename_in_repository(
                package_name, distro, series, arch_tag)

    def test_package_in_ubuntu(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad()

        result = self._verify_package(lp_api, 'mypkg', 'ubuntu', 'lucid',
            'i386')

        ubuntu.getArchive.assert_called_with(name='primary')
        archive.getPublishedBinaries.assert_called_with(
            binary_name='mypkg',
            distro_arch_series='http://lpapi/ubuntu/lucid/i386',
            status='Published', exact_match=True)
        self.assertTrue(result)

    def test_package_not_in_ubuntu(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad(
            binary_name='other_pkg')

        result = self._verify_package(lp_api, 'mypkg', 'ubuntu', 'lucid',
            'i386')

        ubuntu.getArchive.assert_called_with(name='primary')
        archive.getPublishedBinaries.assert_called_with(
            binary_name='mypkg',
            distro_arch_series='http://lpapi/ubuntu/lucid/i386',
            status='Published', exact_match=True)
        self.assertFalse(result)

    def test_package_in_partner(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad()

        result = self._verify_package(lp_api, 'mypkg', 'canonical', 'lucid',
            'i386')

        ubuntu.getArchive.assert_called_with(name='partner')
        archive.getPublishedBinaries.assert_called_with(
            binary_name='mypkg',
            distro_arch_series='http://lpapi/ubuntu/lucid/i386',
            status='Published', exact_match=True)
        self.assertTrue(result)

    def test_package_in_extras(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad()

        result = self._verify_package(lp_api, 'mypkg',
            'lp-ppa-app-review-board', 'lucid', 'i386')

        app_review_board = lp_api.people['app-review-board']
        app_review_board.getPPAByName.assert_called_with(name='ppa')
        archive.getPublishedBinaries.assert_called_with(
            binary_name='mypkg',
            distro_arch_series='http://lpapi/ubuntu/lucid/i386',
            status='Published', exact_match=True)
        self.assertTrue(result)

    def test_unknown_origin(self):
        lp_api, ubuntu, archive = self._make_mock_launchpad()

        result = self._verify_package(lp_api, 'mypkg',
            'lp-ppa-unknown', 'lucid', 'i386')

        # Currently we only support primary, partner, extras and the
        # specified commercial origins.
        self.assertEqual(0, archive.getPublishedBinaries.call_count)
        self.assertFalse(result)

    def test_invalid_series_or_arch(self):
        # An invalid series or arch raises a 400 Bad Request.
        lp_api, ubuntu, archive = self._make_mock_launchpad()
        response = Mock()
        response.status = 400
        archive.getPublishedBinaries.side_effect = HTTPError(
            response, 'content')

        result = self._verify_package(lp_api, 'mypkg', 'ubuntu', 'lucid',
            'i386')

        self.assertFalse(result)

    @patch('reviewsapp.utilities.Launchpad')
    def test_doesnt_sign_in_each_time(self, mock_launchpad):
        lp_api, ubuntu, archive = self._make_mock_launchpad()
        mock_launchpad.login_anonymously.return_value = lp_api
        ws = WebServices()
        ws.lp_verify_packagename_in_repository("foobar", 'ubuntu',
                                               "lucid", 'i386')
        self.assertTrue(mock_launchpad.login_anonymously.called)
        mock_launchpad.login_anonymously.reset_mock()

        ws.lp_verify_packagename_in_repository("foobar", 'ubuntu',
                                               "lucid", 'i386')

        self.assertFalse(mock_launchpad.login_anonymously.called)

    # Patch at a lower level to ensure that the request will be made
    @patch('lazr.restfulclient._browser.Browser._request')
    def test_doesnt_fail_with_home_undefined(self, mock_request):
        mock_request.return_value = {}, ''
        if 'HOME' in os.environ:
            del os.environ['HOME']
        ws = WebServices()
        # SyntaxError is raised when we fail to parse our dummy wadl
        self.assertRaises(SyntaxError, ws.lp_verify_packagename_in_repository,
            "foobar", "ubuntu", "lucid", 'i386')


class SCATestCase(TestCaseWithFactory):

    def _request_sca_pubs_for(self, distroseries, arch_tag, disable_cache=True):
        with patch('urllib.urlretrieve') as mock_urlretrieve:
            mock_urlretrieve.return_value = (
                self.factory.get_test_file('software_center_apps.json'), None)
            if disable_cache:
                with patch('django.core.cache') as mock_cache:
                    mock_cache.get.return_value = False
                    published_apps = WebServices().get_sca_publishings_for(
                        distroseries, arch_tag)
            else:
                published_apps = WebServices().get_sca_publishings_for(
                    distroseries, arch_tag)
        return published_apps, mock_urlretrieve

    def test_get_sca_publishings(self):
        published_apps, ignored = self._request_sca_pubs_for(
            'maverick', 'i386')

        self.assertEqual({
            'lp-ppa-commercial-ppa-uploaders-brukkon': 'brukkon',
            'lp-ppa-commercial-ppa-uploaders-fluendo-dvd': 'fluendo-dvd',
            'lp-ppa-commercial-ppa-uploaders-fluendo-plugins':
                'gstreamer0.10-fluendo-plugins',
            'lp-ppa-commercial-ppa-uploaders-fluendo-wmv-plugins':
                'gstreamer0.10-fluendo-plugins-wmv',
            'lp-ppa-commercial-ppa-uploaders-illumination': 'illumination',
            'lp-ppa-commercial-ppa-uploaders-vendetta-online': 'vendetta-online',
            'lp-ppa-commercial-ppa-uploaders-world-of-goo': 'worldofgoo',
            }, published_apps)

    def test_retrieved_url(self):
        with patch_settings(SCA_HOST_URL='http://example.com/'):
            ignored, mock_urlretrieve = self._request_sca_pubs_for(
                'natty', 'amd64')

        mock_urlretrieve.assert_called_with(
            'http://example.com/apps/en/ubuntu/natty/amd64/')

    def test_get_sca_publishings_cached(self):
        published_apps, mock_url_retrieve = self._request_sca_pubs_for(
            'maverick', 'i386', disable_cache=False)
        published_apps, mock_urlretrieve = self._request_sca_pubs_for(
            'maverick', 'i386', disable_cache=False)

        self.assertEqual(0, mock_urlretrieve.call_count)

    def _request_sca_verify(self, pkgname, origin, distroseries, arch_tag):
        with patch('urllib.urlretrieve') as mock_urlretrieve:
            mock_urlretrieve.return_value = (
                self.factory.get_test_file('software_center_apps.json'), None)
            ws = WebServices()
            return ws.sca_verify_packagename_in_repository(
                pkgname, origin, distroseries, arch_tag)

    def test_sca_verify_packagename(self):
        self.assertTrue(self._request_sca_verify(
            'brukkon', 'lp-ppa-commercial-ppa-uploaders-brukkon',
            'maverick', 'i386'))
        self.assertFalse(self._request_sca_verify(
            'fluendo-dvd', 'lp-ppa-commercial-ppa-uploaders-brukkon',
            'maverick', 'i386'))
        self.assertFalse(self._request_sca_verify(
            'brukkon', 'lp-ppa-commercial-ppa-uploaders-doesnt-exist',
            'maverick', 'i386'))


class InvalidatePaginatedReviewsTestCase(TestCaseWithFactory):
    def _url_for_package(self, package_name, origin='ubuntu',
        distro_series='lucid', language='en', version='1.0', page=None):
        kwargs = {
            'language': language,
            'origin': origin,
            'distroseries': distro_series,
            'version': version,
            'package_name': urllib.quote_plus(package_name),
            }
        if page is not None:
            kwargs['page'] = page

        return reverse('rnr-api-reviews-for-package', kwargs=kwargs)

    def _request_reviews_for_package(self, package_name, origin='ubuntu',
        distro_series='lucid', language='en', version='1.0', page=None):

        url = self._url_for_package(package_name, origin, distro_series,
            language, version, page)

        with patch_settings(API_REVIEWS_PAGINATION=2):
            response = self.client.get(url)

        return response

    def _populate_reviews_and_cache(self, package_name, origin='ubuntu',
        distro_series='lucid', language='en', version='1.0', pages=1):
        package = self.factory.makeSoftwareItem(package_name)
        repository = self.factory.makeRepository(
            origin=origin, distro_series=distro_series)

        # We ensure that API_REVIEWS_PAGINATION is set to 2 reviews per
        # page for our tests.
        reviews = []
        for review_num in range(pages * 2):
            reviews.append(
                self.factory.makeReview(software_item=package,
                    language=language, repository=repository,
                    version=version))

        # Now ensure the cache is populated by requesting each url.
        self._request_reviews_for_package(package_name)
        for page_num in range(1, pages + 1):
            self._request_reviews_for_package(package_name, page=page_num)

        return reviews

    def test_invalidates_all_pages(self):
        # The cache is invalidated for all pages.
        reviews = self._populate_reviews_and_cache('package_foo', pages=3)
        for page in (None, 1, 2, 3):
            url = self._url_for_package('package_foo', page=page)
            self.assertTrue(cache.has_key(cache_key_for_url(url)))

        invalidate_paginated_reviews_for(reviews[0])

        for page in (None, 1, 2, 3):
            url = self._url_for_package('package_foo', page=page)
            self.assertFalse(cache.has_key(cache_key_for_url(url)))

    def test_invalidates_with_space(self):
        # The cache is invalidated for all pages.
        reviews = self._populate_reviews_and_cache('package;App with space')
        invalidate_paginated_reviews_for(reviews[0])
        url = self._url_for_package('package;App with space')
        self.assertFalse(cache.has_key(cache_key_for_url(url)))

    def test_invalidates_single_review_request(self):
        review = self.factory.makeReview()
        url = reverse('rnr-api-review-for-review-id', args=[review.id])
        response = self.client.get(url)
        self.assertTrue(cache.has_key(cache_key_for_url(url)))

        invalidate_paginated_reviews_for(review)

        self.assertFalse(cache.has_key(cache_key_for_url(url)))


class FullClaimedIdTestCase(TestCaseWithFactory):
    def test_full_claimed_id(self):
        """Simple test to check that the right claimed_id is calculated"""
        consumer_key = 'abcdef12'

        claimed_id = full_claimed_id(consumer_key)

        self.assertTrue(claimed_id.startswith(settings.OPENID_SSO_SERVER_URL))
        self.assertTrue(claimed_id.endswith(consumer_key))
