#
#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Ubuntu App Review view tests."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'OverviewTestCase',
    'ReviewModerationModerateViewTestCase',
    'ReviewModerationDetailsTestCase',
    'ReviewModerationListViewTestCase',
    'ArtificialOopsTestCase',
    ]


from django.contrib.auth.models import Permission
from django.core.urlresolvers import reverse

import reviewsapp.urls
from reviewsapp.forms import (
    ReviewModerationFilterForm,
    ReviewModerationModerateForm,
    )
from reviewsapp.models import ReviewModeration
from reviewsapp.tests.helpers import patch_settings
from reviewsapp.tests.factory import TestCaseWithFactory


class OverviewTestCase(TestCaseWithFactory):
    """Tests for the overview page."""

    def test_anonymous_ok(self):
        # Anonymous users can access the page.
        response = self.client.get(reverse('rnr-overview'))

        self.assertTemplateUsed(response, 'reviewsapp/overview.html')

    def test_includes_link_to_moderation(self):
        response = self.client.get(reverse('rnr-overview'))

        url = reverse('rnr-reviewmoderation-list')
        self.assertContains(response, '<a href="{0}"'.format(url))


class ReviewModerationListViewTestCase(TestCaseWithFactory):
    """Tests for the reviewmoderation_list view."""

    def test_moderation_list_without_permission(self):
        # A 302 is returned when trying to view with an unauthorized User
        self.client.login(
            username=self.factory.makeUser().username, password='test')
        review_moderation = self.factory.makeReviewModeration()

        response = self.client.get(reverse('rnr-reviewmoderation-list'))

        self.assertEqual(302, response.status_code)
        self.assertTrue('/reviews/forbidden' in response['Location'])

    def test_moderation_list_anonymous(self):
        # A 302 is returned when trying to view with an anonymous User
        review_moderation = self.factory.makeReviewModeration()

        response = self.client.get(reverse('rnr-reviewmoderation-list'))

        self.assertEqual(302, response.status_code)
        self.assertTrue('/openid/login' in response['Location'])

    def _request_url_as_moderator(self, data=None, user=None):
        # Helper to request the url as a moderator.
        data = data or {}
        if user is None:
            user = self.factory.makeUser()
        can_moderate = Permission.objects.get(codename='can_moderate')
        user.user_permissions.add(can_moderate)
        self.client.login(username=user.username, password='test')
        url = reverse('rnr-reviewmoderation-list')
        return self.client.get(url, data=data)

    def test_filter_by_language(self):
        # Results can be filtered for a language.
        rm_en = self.factory.makeReviewModeration(language="en")
        rm_de = self.factory.makeReviewModeration(language="de")

        response = self._request_url_as_moderator(data={
            'language': 'en',
            })

        review_moderations = response.context['review_moderations']
        self.assertEqual([rm_en], list(review_moderations))

    def test_filter_by_distroseries(self):
        # Results can be filtered for a distroseries.
        rm_hoary = self.factory.makeReviewModeration(distro_series="hoary")
        rm_mav = self.factory.makeReviewModeration(distro_series="maverick")

        response = self._request_url_as_moderator(data={
            'distro_series': 'maverick',
            })

        review_moderations = response.context['review_moderations']
        self.assertEqual([rm_mav], list(review_moderations))

    def test_filter_by_status(self):
        # Results can be filtered for a moderation status.
        rm_pending = self.factory.makeReviewModeration()
        rm_approved = self.factory.makeReviewModeration(
            status=ReviewModeration.APPROVED_STATUS)

        response = self._request_url_as_moderator(data={
            'status': ReviewModeration.APPROVED_STATUS,
            })

        review_moderations = response.context['review_moderations']
        self.assertEqual([rm_approved], list(review_moderations))

    def test_results_ordered_by_flag_count(self):
        # Results are ordered by the flag count.
        review = self.factory.makeReview(num_flags=3, summary="3 flags")
        self.factory.makeReview(num_flags=1, summary="1 flag",
            language=review.language, repository=review.repository)
        self.factory.makeReview(num_flags=8, summary="8 flags",
            language=review.language, repository=review.repository)

        response = self._request_url_as_moderator()

        summaries = [
            review_moderation.review.summary for
                review_moderation in response.context['review_moderations']]
        self.assertEqual(
            ['8 flags', '3 flags', '1 flag'], summaries)

    def test_request_not_cached(self):
        # Requests for server status are never cached.
        review_moderation = self.factory.makeReviewModeration()

        response = self._request_url_as_moderator()

        self.assertEqual('max-age=0', response['cache-control'])

    def test_provides_language_choices(self):
        # The view provides a dynamic choice list for languages.
        self.factory.makeReviewModeration(language='en')
        self.factory.makeReviewModeration(language='en')
        self.factory.makeReviewModeration(language='de')

        response = self._request_url_as_moderator()

        form = response.context['filter_form']
        self.assertTrue(isinstance(form, ReviewModerationFilterForm))
        self.assertEqual(
            [('', 'All languages'), ('de', 'de (1)'), ('en', 'en (2)')],
            form.fields['language'].choices)

    def test_includes_total_items_for_moderation(self):
        # The context includes the total items for moderation (ie. not
        # dependent on the filtering options).
        self.factory.makeReviewModeration(language='en')
        self.factory.makeReviewModeration(language='en')
        self.factory.makeReviewModeration(language='de')

        response = self._request_url_as_moderator(data={
            'language': 'en',
            })

        self.assertEqual(2, response.context['total'])

    def test_includes_linked_user(self):
        # The reviewer is marked-up correctly.
        # This is really ensuring that the lp_user template filter is
        # working and should be replaced with proper tests of the filter
        # itself.
        moderation = self.factory.makeReviewModeration(language='en')

        response = self._request_url_as_moderator()

        reviewer = moderation.review.reviewer
        self.assertContains(response,
            '<a href="https://launchpad.net/~{0}">{1}</a>'.format(
                reviewer.username, reviewer.get_full_name()))

    def test_non_ascii_reviewer(self):
        moderation = self.factory.makeReviewModeration(language='en')
        reviewer = moderation.review.reviewer
        reviewer.username = u'\xa1'
        reviewer.first_name = u'\xa2'
        reviewer.last_name = u'\xa3'
        reviewer.save()

        response = self._request_url_as_moderator()

        self.assertContains(response,
            u'<a href="https://launchpad.net/~{0}">{1}</a>'.format(
                reviewer.username, reviewer.get_full_name()))


    def test_includes_next_url_with_get_params(self):
        # The next_url inludes current filtering options.
        response = self._request_url_as_moderator(data=dict(language='en'))

        url = reverse('rnr-reviewmoderation-list') + "?language=en"
        self.assertEqual(url, response.context['next_url'])

    def test_includes_moderate_form(self):
        # A form is included for each moderation item.
        moderation = self.factory.makeReviewModeration()

        response = self._request_url_as_moderator()

        action_url = reverse('rnr-reviewmoderation-moderate',
            args=[moderation.id])
        self.assertContains(response,
            '<form method="post" action="{0}"'.format(action_url))

    def test_no_form_when_user_cant_moderate(self):
        # The form is not included if the user is not allowed to
        # moderate an item.
        moderation = self.factory.makeReviewModeration()

        response = self._request_url_as_moderator(
            user=moderation.review.reviewer)

        action_url = reverse('rnr-reviewmoderation-moderate',
            args=[moderation.id])
        self.assertNotContains(response, action_url)

    def test_displays_when_flagger_is_none(self):
        # A form is included for each moderation item.
        review = self.factory.makeReview()
        review.flag(None, 'summary', 'description')
        moderation = review.reviewmoderation_set.all()[0]

        response = self._request_url_as_moderator()

        action_url = reverse('rnr-reviewmoderation-moderate',
            args=[moderation.id])
        self.assertContains(response,
            '<form method="post" action="{0}"'.format(action_url))


class ReviewModerationModerateViewTestCase(TestCaseWithFactory):
    """Tests for the moderate view methed."""

    def test_moderation_without_permission(self):
        # A 302 is returned when trying to view with an unauthorized User
        # redirecting to a forbidden explanation page (rather than doing
        # an openid redirect loop).
        self.client.login(
            username=self.factory.makeUser().username, password='test')
        review_moderation = self.factory.makeReviewModeration()

        url = reverse('rnr-reviewmoderation-moderate',
                      args=[review_moderation.id])
        response = self.client.get(url)

        self.assertRedirects(
            response, '/reviews/forbidden?next=' + url)

    def test_moderation_anonymous(self):
        # A 302 is returned when trying to view with an anonymous User
        # redirecting to the login page.
        review_moderation = self.factory.makeReviewModeration()

        url = reverse('rnr-reviewmoderation-moderate',
                      args=[review_moderation.id])
        response = self.client.get(url)

        self.assertEqual(302, response.status_code)
        self.assertEqual(
            'http://testserver/openid/login/?next=' + url,
            response['Location'])

    def _request_moderation_url(self, id, user=None, post=False, data={}):
        # Helper to request the url for the given review_moderation as a
        # moderator.
        if user is None:
            user = self.factory.makeUser()
        can_moderate = Permission.objects.get(codename='can_moderate')
        user.user_permissions.add(can_moderate)
        self.client.login(
            username=user.username, password='test')
        url = reverse('rnr-reviewmoderation-moderate', args=[id])
        if post:
            response = self.client.post(url, data=data)
        else:
            response = self.client.get(url, data=data)

        return response

    def test_review_moderation_doesnt_exist(self):
        # A 404 is returned when the review_moderation doesn't exist.
        response = self._request_moderation_url(999)

        self.assertEqual(404, response.status_code)

    def test_get_nonpending_moderation_redirects_to_details(self):
        # A redirect to the moderation detail results if not PENDING.
        review_moderation = self.factory.makeReviewModeration(
            status=ReviewModeration.APPROVED_STATUS)

        response = self._request_moderation_url(review_moderation.id)

        self.assertEqual(302, response.status_code)
        url = reverse('rnr-reviewmoderation-detail',
            args=[review_moderation.id])
        self.assertEqual('http://testserver' + url, response['Location'])

    def test_get_pending_moderation_includes_form(self):
        # A form to update the moderation is included when pending.
        review_moderation = self.factory.makeReviewModeration()

        response = self._request_moderation_url(review_moderation.id)

        self.assertTemplateUsed(
            response, 'reviewsapp/reviewmoderation_form.html')
        self.assertIsInstance(
            response.context['form'], ReviewModerationModerateForm)
        self.assertContains(
            response, '<form method="post" action="{0}">'.format(
                reverse('rnr-reviewmoderation-moderate',
                        args=[review_moderation.id])))

    def test_invalid_post_includes_form(self):
        # An invalid post results in the re-rendering of the form with
        # input validation.
        review_moderation = self.factory.makeReviewModeration()

        response = self._request_moderation_url(
            review_moderation.id, post=True, data=dict(
                moderation_text='hello'))

        self.assertTemplateUsed(
            response, 'reviewsapp/reviewmoderation_form.html')
        form = response.context['form']
        self.assertEqual(
            set(form.errors.keys()), set(['status']))

    def test_successful_post_redirects_to_details(self):
        review_moderation = self.factory.makeReviewModeration()

        response = self._request_moderation_url(
            review_moderation.id, post=True, data={
                'moderation_text': 'blah',
                'status': ReviewModeration.APPROVED_STATUS,
                })

        self.assertEqual(302, response.status_code)
        url = reverse('rnr-reviewmoderation-detail',
            args=[review_moderation.id])
        self.assertEqual('http://testserver' + url, response['Location'])

    def test_successful_post_updates_review_moderation(self):
        review_moderation = self.factory.makeReviewModeration()

        response = self._request_moderation_url(
            review_moderation.id, post=True, data={
                'moderation_text': 'blah',
                'status': ReviewModeration.APPROVED_STATUS,
                })

        review_moderation = ReviewModeration.objects.get(
            id=review_moderation.id)
        self.assertEqual(ReviewModeration.APPROVED_STATUS,
                         review_moderation.status)
        self.assertEqual('blah', review_moderation.moderation_text)

    def test_redirects_to_next(self):
        # If a next parameter is included in the post, we redirect
        # there.
        review_moderation = self.factory.makeReviewModeration()

        response = self._request_moderation_url(
            review_moderation.id, post=True, data={
                'moderation_text': 'blah',
                'status': ReviewModeration.APPROVED_STATUS,
                'next_url': 'http://example.com/',
                })

        self.assertEqual(302, response.status_code)
        self.assertEqual('http://example.com/', response['Location'])

    def test_creates_message_on_moderation(self):
        # A notification is created on moderation.
        review_moderation = self.factory.makeReviewModeration()
        user = self.factory.makeUser()
        self.assertEqual(0, user.message_set.count())

        response = self._request_moderation_url(
            review_moderation.id, user=user, post=True, data={
                'moderation_text': 'blah',
                'status': ReviewModeration.APPROVED_STATUS,
                })

        self.assertEqual(1, user.message_set.count())

    def test_cannot_moderate_gives_403(self):
        # A 405 is returned if user is not able to moderate this
        # specific moderation.
        review_moderation = self.factory.makeReviewModeration()
        reviewer = review_moderation.review.reviewer

        response = self._request_moderation_url(
            review_moderation.id, user=reviewer, post=True, data={
                'moderation_text': 'blah',
                'status': ReviewModeration.APPROVED_STATUS,
                })

        self.assertEqual(403, response.status_code)


class ReviewModerationDetailsTestCase(TestCaseWithFactory):
    """Tests for the show_moderation_details view method."""

    def _request_moderation_detail(self, id):
        user = self.factory.makeUser()
        can_moderate = Permission.objects.get(codename='can_moderate')
        user.user_permissions.add(can_moderate)
        self.client.login(
            username=user.username, password='test')
        url = reverse('rnr-reviewmoderation-detail', args=[id])
        return self.client.get(url)

    def test_moderation_details(self):
        # Details for a specific moderation request can be displayed.
        review_moderation = self.factory.makeReviewModeration()
        review = review_moderation.review

        response = self._request_moderation_detail(review_moderation.id)

        self.assertTemplateUsed(
            response, 'reviewsapp/reviewmoderation_details.html')
        # The title/h1 share the same overview text for the package.
        self.assertContains(
            response,
            'Pending moderation for package {0}'.format(
                review.softwareitem.package_name), count=2)
        # The flags summary and description are included.
        flag = review_moderation.reviewmoderationflag_set.all()[0]
        self.assertContains(response, flag.summary, count=1)
        self.assertContains(
            response, flag.description, count=1)

    def test_missing_moderation_details(self):
        # A 404 occurs when the requested moderation does not exist.
        response = self._request_moderation_detail(999)

        self.assertEqual(response.status_code, 404)

    def test_forbidden_moderation_details(self):
        # A 302 occurs when the user doesn't have the right permissions.
        self.client.login(
            username=self.factory.makeUser().username, password='test')

        url = reverse('rnr-reviewmoderation-detail', args=[1])
        response = self.client.get(url)

        self.assertEqual(302, response.status_code)
        self.assertTrue('/reviews/forbidden' in response['Location'])

    def test_anonymous_moderation_details(self):
        # A 302 occurs when requesting with an AnonymousUser
        url = reverse('rnr-reviewmoderation-detail', args=[1])
        response = self.client.get(url)

        self.assertEqual(302, response.status_code)
        self.assertTrue('/openid/login' in response['Location'])


class ArtificialOopsTestCase(TestCaseWithFactory):
    def test_disabling_removes_url(self):
        with patch_settings(ENABLE_ARTIFICIAL_OOPS=False):
            # We need to reload urls.py to recalculate all urls
            reload(reviewsapp.urls)
            
            response = self.client.get('/reviews/die/')

        self.assertEquals(404, response.status_code)
        reload(reviewsapp.urls)

    def test_die(self):
        with patch_settings(ENABLE_ARTIFICIAL_OOPS=True):
            # We need to reload urls.py to recalculate all urls
            reload(reviewsapp.urls)
            
            self.assertRaises(ZeroDivisionError, self.client.get,
                '/reviews/die/')

        reload(reviewsapp.urls)
