import os

from tempfile import mkdtemp
from shutil import rmtree

from django.conf import settings
from django.core.urlresolvers import reverse
from django.utils.http import urlencode
from mock import Mock, patch

from reviewsapp.wsgi import make_app
from reviewsapp.tests.helpers import patch_settings
from reviewsapp.tests.factory import TestCaseWithFactory


class WSGIDispatchTestCase(TestCaseWithFactory):

    def setUp(self):
        super(WSGIDispatchTestCase, self).setUp()
        """Create a temporary directory for oopses."""
        self.oops_dir = mkdtemp()

        self.environ = {
            'PATH_INFO':         '/',
            'QUERY_STRING':      '',
            'REMOTE_ADDR':       '127.0.0.1',
            'REQUEST_METHOD':    'GET',
            'SCRIPT_NAME':       '',
            'SERVER_NAME':       'testserver',
            'SERVER_PORT':       '80',
            'SERVER_PROTOCOL':   'HTTP/1.1',
            'wsgi.version':      (1, 0),
            'wsgi.url_scheme':   'http',
            'wsgi.multiprocess': True,
            'wsgi.multithread':  False,
            'wsgi.run_once':     False,
        }
        self.orig_middleware = settings.MIDDLEWARE_CLASSES
        # Ensure the LogExceptionMiddleware is enabled, and caching disabled:
        settings.MIDDLEWARE_CLASSES = [
            'django.middleware.common.CommonMiddleware',
            'reviewsapp.middleware.exception.LogExceptionMiddleware',
        ]


    def tearDown(self):
        super(WSGIDispatchTestCase, self).tearDown()
        """Delete temporary oops directory."""
        rmtree(self.oops_dir)
        settings.MIDDLEWARE_CLASSES = self.orig_middleware

    def request_404_exception(self, data={}):
        """Helper for creating a 404 exception with a referrer."""
        self.environ['PATH_INFO'] = '/reviews/made_up_url'
        self.environ['HTTP_REFERER'] = '/'
        if data != {}:
            self.environ['QUERY_STRING'] = urlencode(data, doseq=True)

        response = self.request_wsgi_response(self.environ)
        return response

    def request_500_exception(self, data={}):
        """Helper for creating a 500 exception."""
        self.environ['PATH_INFO'] = reverse('rnr-overview')
        if data != {}:
            self.environ['QUERY_STRING'] = urlencode(data, doseq=True)

        filter_method = 'reviewsapp.models.ReviewModeration.objects.filter'
        with patch(filter_method) as mock_filter:
            mock_filter.side_effect = ZeroDivisionError
            response = self.request_wsgi_response(self.environ)
        return response

    def request_wsgi_response(self, environ):
        start_response = Mock()
        with patch_settings(OOPS_DIR=self.oops_dir):
            app = make_app()
            response = app(environ, start_response).next()
        return response

    def read_oops_from_disk(self):
        # Helper to read in the oops written to disk.
        # There is only one date directory in the oops directory.
        dated_dirs = os.listdir(self.oops_dir)
        self.assertEqual(1, len(dated_dirs))

        # There is only one oops.
        todays_oops_path = dated_dirs[0]
        oops_list = os.listdir(self.oops_dir + '/' + todays_oops_path)
        self.assertEqual(1, len(oops_list))

        oops_path = self.oops_dir + '/' + todays_oops_path + '/' + oops_list[0]
        with open(oops_path, "r") as oops_file:
            oops_file_contents = oops_file.read()
        return oops_file_contents

    def test_oopsid_in_404_response(self):
        # The rendered response for a view exception includes the oops id.
        response = self.request_404_exception()

        self.assertTrue(self.environ['OOPSID'] in response)

    def test_oopsid_in_500_response(self):
        # The rendered response for a 500 exception includes the oops id.
        response = self.request_500_exception()

        self.assertTrue(self.environ['OOPSID'] in response)

    def test_exception_in_404_oops_on_disk(self):
        # A request resulting in an 404 with a referrer writes an oops to disk
        response = self.request_404_exception()

        self.assertTrue("Exception-Type:" in self.read_oops_from_disk())

    def test_traceback_in_500_oops_on_disk(self):
        # A request resulting in a 500 writes the oops to disk.
        response = self.request_500_exception()

        # The oops contains the stacktrace as part of a log message.
        self.assertTrue(
            "'message': 'Traceback (most recent call last):" in
                self.read_oops_from_disk())
