#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SSO/OAuth authenticator for Piston."""


from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'cache_key_for_url',
    'WebServices',
    'WebServiceError',
    ]

import logging
import urllib
from httplib2 import ServerNotFoundError

from django.conf import settings
from django.core.cache import cache
from django.core.urlresolvers import reverse
from django.http import HttpRequest
from django.utils.cache import get_cache_key
from django.utils.http import urlquote_plus
from django.utils import simplejson

from launchpadlib.launchpad import Launchpad
from launchpadlib.uris import lookup_service_root
from lazr.restfulclient.authorize import BasicHttpAuthorizer
from lazr.restfulclient.authorize.oauth import OAuthAuthorizer
from lazr.restfulclient.errors import HTTPError
from lazr.restfulclient.resource import ServiceRoot
from oauth.oauth import OAuthToken

from reviewsapp.models import Review


class WebServiceError(Exception):
    """Raised when we cannot connect to a web service."""


class WebServices:
    _identity_provider = None
    _launchpad_service = None

    def __init__(self):
        """Pre-create the identity provider webservice if it will be used."""
        if settings.PRELOAD_API_SERVICE_ROOTS:
            self._set_identity_provider()
            self._set_launchpad_service()

    @property
    def logger(self):
        return logging.getLogger('rnr.utilities')

    def _create_service(self, authorizer, service_root_url):
        try:
            service_root = ServiceRoot(authorizer, service_root_url)
            self.logger.info(
                "Webservice root creation successful: %s" % service_root_url)
            return service_root
        except (HTTPError, ServerNotFoundError), e:
            self.logger.exception(
                "Failed to create service root: %s" % service_root_url)
            return None

    def _create_basic_auth_service(self, service_root_url, username, password):
        authorizer = BasicHttpAuthorizer(username, password)
        return self._create_service(authorizer, service_root_url)

    def _set_identity_provider(self):
        WebServices._identity_provider = self._create_basic_auth_service(
            settings.SSO_API_SERVICE_ROOT, settings.SSO_API_AUTH_USERNAME,
            settings.SSO_API_AUTH_PASSWORD)

    def _set_launchpad_service(self):
        service_root = lookup_service_root(settings.LP_SERVICE)
        launchpadlib_dir = settings.LAUNCHPADLIB_DIR
        try:
            WebServices._launchpad_service = Launchpad.login_anonymously(
                settings.LP_CONSUMER_NAME, service_root,
                launchpadlib_dir=launchpadlib_dir)
        except (HTTPError, ServerNotFoundError), e:
            self.logger.exception('Unable to create Launchpad API.')
            WebServices._launchpad_service = None

    @property
    def identity_provider(self):
        if self._identity_provider is None:
            self._set_identity_provider()

            if self._identity_provider is None:
                raise WebServiceError("The payment service is not available.")

        return self._identity_provider

    @property
    def launchpad_service(self):
        if self._launchpad_service is None:
            self._set_launchpad_service()

        if self._launchpad_service is None:
            raise WebServiceError("The Launchpad service is not available.")

        return self._launchpad_service

    def _fake_validate_token(self, token, openid_identifier, signature):
        """ This is a version of validate_token that gets the
            token_secret, consumer_secret from the plaintext signature.

            It will break once we use a real signature for oauth, but
            its very useful for testing as it does not require the special
            priviledges required for the call to
            identity_provider.authentications.validate_token()

            The token can still be validated (and will be) with a
            api.accounts.me() call
        """
        (consumer_secret, token_secret) = urllib.unquote(signature).split("&")
        result = {}
        result["token"] = token
        result["token_secret"] = token_secret
        result["consumer_key"] = openid_identifier
        result["consumer_secret"] = consumer_secret
        result["name"] = "fake-token"
        return result

    def get_data_for_account(self, token, openid_identifier, signature=None):
        if settings.SSO_AUTH_MODE_NO_UBUNTU_SSO_PLAINTEXT_ONLY:
            # token extraction based on PLAINTEXT token, the function
            # "validate_token" is really a "get_secrets_for_token()" call
            result = self._fake_validate_token(
                token, openid_identifier, signature)
        else:
            result = self.identity_provider.authentications.validate_token(
                token=token, consumer_key=openid_identifier)

        if result:
            oauth_token = OAuthToken(result['token'], result['token_secret'])
            authorizer = OAuthAuthorizer(result['consumer_key'],
                result['consumer_secret'], oauth_token)
            api = self._create_service(authorizer,
                settings.SSO_API_SERVICE_ROOT)
            result.update(api.accounts.me())
        return result

    def lp_verify_packagename_in_repository(self, pkgname, origin, distroseries,
                                            arch_tag):
        """Verify that a given package exists in a distroseries on LP.

        :param pkgname: The name of the package to check.
        :param origin: The repository origin identifier.
        :param distroseries: The name of distribution series.
        :param arch_tag: The machine architecture (i386, amd64...)
        :return: Whether the package was found or not.
        :rtype: bool
        """
        ubuntu = self.launchpad_service.distributions['ubuntu']
        if origin == 'ubuntu':
            archive = ubuntu.getArchive(name='primary')
        elif origin == 'canonical':
            archive = ubuntu.getArchive(name='partner')
        elif origin =='lp-ppa-app-review-board':
            app_review_board = self.launchpad_service.people[
                'app-review-board']
            archive = app_review_board.getPPAByName(name='ppa')
        else:
            return False

        # We now have the archive, check for the distroseries and the
        # package.
        distro_arch_series = ubuntu.self_link + '/{0}/{1}'.format(
            distroseries, arch_tag)
        try:
            binaries = archive.getPublishedBinaries(
                binary_name=pkgname, distro_arch_series=distro_arch_series,
                status='Published', exact_match=True)
        except HTTPError as error:
            if error.response.status == 400:
                # A bad request is raised if the distro arch series we
                # indicate does not exist.  Obviously, this does not
                # verify!  Any other status is unexpected.
                return False
            raise
        for binary in binaries:
            if binary.binary_package_name == pkgname:
                return True
        return False

    def sca_verify_packagename_in_repository(self, pkgname, origin,
                                             distroseries, arch_tag):
        """Verify that a given package exists in a distroseries on SCA.

        :param pkgname: The name of the package to check.
        :param origin: The repository origin identifier.
        :param distroseries: The name of distribution series.
        :param arch_tag: The machine architecture (i386, amd64...)
        :return: Whether the package was found or not.
        :rtype: bool
        """
        sca_publishings = self.get_sca_publishings_for(distroseries, arch_tag)
        if origin in sca_publishings:
            # XXX 2011-03-11 michaeln bug=733170 PPA origins are
            # currently ambiguous.
            published_apps = self.get_sca_publishings_for(distroseries, arch_tag)
            return pkgname == published_apps[origin]

        return False

    def get_sca_publishings_for(self, distroseries, arch_tag):
        """A helper that builds a dict of publishings for sca-sold apps.

        XXX michaeln 2011-03-14 bug=734747 Optional parms for available_apps.
        It would be great if we could just pull /apps/en/ubuntu/ and use
        each items 'series' dict to populate this, rather than pulling each
        distroserie/arch separately. If we can assume that an app sold via
        softwarecenter will always be available for maverick i386, then we
        could do just the one call too.
        """
        cache_key = 'sca_{0}_{1}'.format(distroseries, arch_tag)
        result = cache.get(cache_key)
        if result is not None:
            return result

        # XXX michaeln 2011-03-14 Update to use sca piston client.
        # Once production SCA includes the piston client version we can
        # use that instead. At time of writing the following url works
        # on staging but not production sca:
        #     api/2.0/applications/en/ubuntu/maverick/i386/
        filename, headers = urllib.urlretrieve(
            settings.SCA_HOST_URL +
            'apps/en/ubuntu/{0}/{1}/'.format(distroseries, arch_tag))

        with open(filename, 'r') as sca_apps_file:
            sca_apps = simplejson.load(sca_apps_file)

        # We ensure the keys match Launchpad's current (ambiguous)
        # origin identifiers - Launchpad bug 733170.
        app_names = dict([(
            'lp-ppa-' + app['archive_id'].replace('/', '-'),
            app['package_name'],
            ) for app in sca_apps])
        cache.set(cache_key, app_names, settings.SCA_APPS_CACHE_TIMEOUT)
        return app_names


def cache_key_for_url(url):
    """Return the key used by django's caching system for a (non-varied) url."""
    request = HttpRequest()
    request.path = url
    return get_cache_key(request)


def invalidate_paginated_reviews_for(review):
    # XXX achuni 2011-01-19: Here we're invalidating pages only for the exact
    # version, language, distroseries and origin of the review.  We'd also
    # need to invalidate partial matches, up to including all reviews
    # for the package (/reviews/any/any/any/any/packagename/).  This is
    # a lot of work.  Is it worth it?
    language = review.language
    origin = review.repository.origin
    distroseries = review.repository.distroseries
    package_name = review.softwareitem.package_name
    app_name = review.app_name
    version = review.version

    if app_name:
        package_app_name = u"{0};{1}".format(package_name, app_name)
    else:
        package_app_name = package_name

    num_reviews_for_pkg = Review.objects.filter(
        repository__origin=origin, repository__distroseries=distroseries,
        softwareitem__package_name=package_name,
        app_name=app_name, hide=False, language=language,
        version=version).count()

    # Ensure we've got a list of all the possible page numbers,
    # including None (ie. first page is displayed if there is no page
    # number).
    pages = range(1, num_reviews_for_pkg + 1)
    pages.insert(0, None)
    for page in pages:
        url_kwargs = dict(
            language=language, origin=origin, distroseries=distroseries,
            version=version, package_name=urlquote_plus(package_app_name))
        if page:
            url_kwargs['page'] = page

        package_url = reverse('rnr-api-reviews-for-package', kwargs=url_kwargs)
        key = cache_key_for_url(package_url)
        if cache.has_key(key):
            cache.delete(key)


    # Delete the cache entry for the single review request also.
    url = reverse('rnr-api-review-for-review-id', args=[review.id])
    key = cache_key_for_url(url)
    if cache.has_key(key):
        cache.delete(key)

def full_claimed_id(consumer_key):
    return '%s/+id/%s' % (settings.OPENID_SSO_SERVER_URL.strip('/'),
        consumer_key)
