#   This file is part of Software Center Ratings and Reviews
#   Copyright (C) 2010 Canonical Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as
#   published by the Free Software Foundation, either version 3 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Django views."""

from __future__ import absolute_import

__metaclass__ = type
__all__ = [
    'forbidden',
    'reviewmoderation_list',
    'reviewmoderation_moderate',
    'reviewmoderation_detail',
    ]


from urllib import urlencode

from django import http
from django.db.models import Count
from django.contrib.auth import logout
from django.contrib.auth.decorators import login_required, permission_required
from django.core.urlresolvers import reverse
from django.http import (
    HttpResponseForbidden,
    HttpResponseRedirect,
    )
from django.shortcuts import (
    get_object_or_404,
    render_to_response,
    )
from django.template import loader, RequestContext
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

from reviewsapp.forms import (
    ReviewModerationFilterForm,
    ReviewModerationModerateForm,
    )
from reviewsapp.models import ReviewModeration


@login_required
@permission_required('reviewsapp.can_moderate', login_url='/reviews/forbidden')
def reviewmoderation_list(request):
    """A filterable list of review moderations with a useful default."""

    review_moderations = ReviewModeration.objects.all()
    next_url = reverse('rnr-reviewmoderation-list')

    # By default we'll be displaying pending moderations.
    status = ReviewModeration.PENDING_STATUS
    if request.GET:
        form = ReviewModerationFilterForm(request.GET)
        if form.is_valid():
            language = form.cleaned_data['language']
            if language:
                review_moderations = review_moderations.filter(
                    review__language=language)

            distro_series = form.cleaned_data['distro_series']
            if distro_series:
                review_moderations = review_moderations.filter(
                    review__repository__distroseries=distro_series)

            status = form.cleaned_data['status']

        next_url += "?" + urlencode(request.GET)
    else:
        form = ReviewModerationFilterForm(
            dict(status=ReviewModeration.PENDING_STATUS))

    review_moderations = review_moderations.filter(status=status)


    # By default we order with the highest flag-count first.
    total = review_moderations.count()
    review_moderations = review_moderations.annotate(
        num_flags=Count('reviewmoderationflag')).order_by(
            '-num_flags')

    return render_to_response(
        'reviewsapp/reviewmoderation_list.html', {
            'review_moderations': review_moderations,
            'filter_form': form,
            'total': total,
            'next_url': mark_safe(next_url),
            }, RequestContext(request))


@login_required
@permission_required('reviewsapp.can_moderate', login_url='/reviews/forbidden')
def reviewmoderation_moderate(request, moderation_id):
    """Approve or reject a review.
    :param request: The Django request object.
    :param moderation_id: The database id of the moderation request to view.
    :return: A redirect to the updated review moderation on success, or
        a form with validation info on failure.
    """
    review_moderation = get_object_or_404(
        ReviewModeration, id=moderation_id)

    # Only pending review moderations can be moderated. In case someone
    # else has just approved/rejected a review, we redirect to the
    # review detail.
    if review_moderation.status != ReviewModeration.PENDING_STATUS:
        return HttpResponseRedirect(
            reverse('rnr-reviewmoderation-detail',
                args=[review_moderation.id]))

    if not review_moderation.can_be_moderated_by(request.user):
        return HttpResponseForbidden(
            "You are not allowed to moderate this review.")

    if request.POST:
        form = ReviewModerationModerateForm(request.POST)
        if form.is_valid():
            review_moderation.moderate(
              request.user, form.cleaned_data['status'],
              form.cleaned_data['moderation_text'])
            request.user.message_set.create(
                message=_(
                    "The review of {app_name} by {reviewer} "
                    "is now {status}").format(
                        app_name=review_moderation.review.package_app_name(),
                        reviewer = review_moderation.review.reviewer.get_full_name(),
                        status = review_moderation.get_status_display().lower()))

            next_url = request.POST.get('next_url')
            if not next_url:
                next_url = reverse('rnr-reviewmoderation-detail',
                    args=[review_moderation.id])

            return HttpResponseRedirect(next_url)
    else:
        form = ReviewModerationModerateForm()

    return render_to_response(
        'reviewsapp/reviewmoderation_form.html', {
            'form': form,
            'review_moderation': review_moderation,
            })


@login_required
@permission_required('reviewsapp.can_moderate', login_url='/reviews/forbidden')
def reviewmoderation_detail(request, moderation_id):
    """Show the details of a specific moderation request.

    :param request: The Django request object.
    :param moderation_id: The database id of the moderation request to view.
    :return: A rendered moderation detail or a 404.
    """
    review_moderation = get_object_or_404(
        ReviewModeration, id=moderation_id)
    return render_to_response('reviewsapp/reviewmoderation_details.html', {
        'review_moderation': review_moderation,
        'package_name': review_moderation.review.softwareitem.package_name,
        })


def forbidden(request):
    logout(request)
    args = {}
    if 'next' in request.GET:
        args['next'] = request.GET['next']
    return render_to_response('forbidden.html', args)


class ErrorPage(object):
    def __init__(self, status, errormsg=None):
        self.template_name = '%s.html' % status
        self.status_code = status
        self.errormsg = errormsg

    def __call__(self, request):
        return self.render_error(request)

    def render_error(self, request):
        template = loader.get_template(self.template_name)
        oopsid = request.META.get('OOPSID')
        atts = {
                'oopsid': oopsid,
                'errormsg': self.errormsg,
               }
        context = RequestContext(request, atts)
        return http.HttpResponse(template.render(context),
                                 status=self.status_code)

server_error = ErrorPage(500)
page_not_found = ErrorPage(404)
